/*
COPYRIGHT (C) 1999  Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
*/


#define LINUX_FPU 0
#define SCHED_IPI     RTAI_4_IPI
#define SCHED_VECTOR  RTAI_4_VECTOR

/* if PRIORDs below are not defined msg and sem queues are FIFOs otherwise   */
/* insertion is in priority order and receiving tasks have their priority    */
/* boosted to that of the sending task, if it is higher than their base one  */

#define NSEM_PRIORD
#define MSG_PRIORD

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/timex.h>
#include <linux/sched.h>

#include <asm/param.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>

#define INTERFACE_TO_LINUX
#include <rtai.h>
#include <rtai_mups.h>

#define RT_TASK_MAGIC 0x754d2774

#define RT_SEM_MAGIC 0xaabcdeff

#define SEM_ERR (0xFfff)

#define MSG_ERR ((RT_TASK *)0xFfff)

#define NOTHING ((void *)0)

#define SOMETHING ((void *)1)

#define TIMER_FREQ FREQ_APIC
#define TIMER_LATENCY LATENCY_APIC
#define TIMER_SETUP_TIME SETUP_TIME_APIC

#define rt_exchange_tasks(oldtask, newtask) \
	__asm__ __volatile__( \
	"pushl %%eax\n\t" \
	"pushl %%ebp\n\t" \
	"pushl %%edi\n\t" \
	"pushl %%esi\n\t" \
	"pushl %%edx\n\t" \
	"pushl %%ecx\n\t" \
	"pushl %%ebx\n\t" \
	"pushl $1f\n\t" \
	"movl (%%ecx), %%ebx\n\t" \
	"movl %%esp, (%%ebx)\n\t" \
	"movl (%%edx), %%esp\n\t" \
	"movl %%edx, (%%ecx)\n\t" \
	"ret\n\t" \
"1:	popl %%ebx\n\t \
	popl %%ecx\n\t \
	popl %%edx\n\t \
	popl %%esi\n\t \
	popl %%edi\n\t \
	popl %%ebp\n\t \
	popl %%eax\n\t" \
	: /* no output */ \
	: "c" (&oldtask), "d" (newtask) \
	: "bx", "cx", "dx");

static RT_TASK *rt_smp_current[NR_RT_CPUS];

static RT_TASK *rt_smp_fpu_task[NR_RT_CPUS];

static int rt_smp_linux_cr0[NR_RT_CPUS];

static RT_TASK rt_smp_linux_task[NR_RT_CPUS];

static RTIME rt_smp_time_h[NR_RT_CPUS];

static int rt_smp_half_tick[NR_RT_CPUS];

static int rt_smp_oneshot_timer[NR_RT_CPUS];

static int rt_smp_not_fired[NR_RT_CPUS];

static int rt_smp_preempt_always[NR_RT_CPUS];

static struct rt_times *linux_times;

#define DECLARE_RT_CURRENT int cpuid; RT_TASK *rt_current

#define RT_CURRENT rt_smp_current[hard_cpu_id()]

#define ASSIGN_RT_CURRENT rt_current = rt_smp_current[cpuid = hard_cpu_id()]

#define fpu_task (rt_smp_fpu_task[cpuid])

#define linux_cr0 (rt_smp_linux_cr0[cpuid])

#define rt_linux_task (rt_smp_linux_task[cpuid])

#define rt_time_h (rt_smp_time_h[cpuid])

#define rt_half_tick (rt_smp_half_tick[cpuid])

#define oneshot_timer (rt_smp_oneshot_timer[cpuid])

#define oneshot_timer_cpuid (rt_smp_oneshot_timer[hard_cpu_id()])

#define not_fired (rt_smp_not_fired[cpuid])

#define preempt_always (rt_smp_preempt_always[cpuid])

#define rt_times (rt_smp_times[cpuid])

#define RT_SCHEDULE(task, cpuid) \
	{ \
		if ((task)->runnable_on_cpus != (cpuid)) { \
			send_sched_ipi((task)->runnable_on_cpus); \
		} else { \
			rt_schedule(); \
		} \
	}

#define RT_SCHEDULE_BOTH(task, cpuid) \
	{ \
		if ((task)->runnable_on_cpus != (cpuid)) { \
			send_sched_ipi((task)->runnable_on_cpus); \
		} \
		rt_schedule(); \
	}

/* ++++++++++++++++++++++++++++++++ TASKS ++++++++++++++++++++++++++++++++++ */

static atomic_t scheduling_cpus = ATOMIC_INIT(0);

static inline void sched_get_global_lock(int cpuid)
{
	if (!test_and_set_bit(cpuid, locked_cpus)) {
		while (test_and_set_bit(31, locked_cpus) && 
					!atomic_read(&scheduling_cpus));
	}
	atomic_inc(&scheduling_cpus);
}


static inline void sched_release_global_lock(int cpuid)
{
	if (test_and_clear_bit(cpuid, locked_cpus) && 
				atomic_dec_and_test(&scheduling_cpus)) {
		clear_bit(31, locked_cpus);
	}
}


static inline void send_sched_ipi(unsigned long dest)
{
	apic_write(APIC_ICR2, (apic_read(APIC_ICR2) & 0xFFFFFF) | (dest << 24));
	apic_write(APIC_ICR, (apic_read(APIC_ICR) & ~0xCDFFF) | SCHED_VECTOR);
}


static void rt_startup(void (*rt_thread)(int), int data)
{
	DECLARE_RT_CURRENT;
	ASSIGN_RT_CURRENT;
	rt_global_sti();
	(*rt_thread)(data);
	rt_task_delete(rt_current);
}


int rt_task_init(RT_TASK *task, void (*rt_thread)(int), int data,
			int stack_size, int priority, int uses_fpu,
			void(*signal)(void), unsigned int run_on_cpu)
{
	int *st;
	unsigned long flags;
	int cpuid;

	if (task->magic == RT_TASK_MAGIC) {
		return -EINVAL;
	}

	st = (int *) kmalloc(stack_size, GFP_KERNEL);   
	if (!st) {
		return -ENOMEM;
	}

	if (run_on_cpu >= NR_RT_CPUS) {
		return -EINVAL;
	}

	task->stack = st + stack_size / sizeof(int);    
	task->uses_fpu = uses_fpu;
	task->runnable_on_cpus = run_on_cpu;
	task->stack_bottom = st;
	task->magic = RT_TASK_MAGIC; 
	task->state = (SUSPENDED | READY);
	task->priority = task->base_priority = priority;
	task->period = 0;
	task->resume_time = RT_TIME_END;
	task->queue.prev = &(task->queue);      
	task->queue.next = &(task->queue);      
	task->queue.task = task;
	task->msg_queue.prev = &(task->msg_queue);      
	task->msg_queue.next = &(task->msg_queue);      
	task->msg_queue.task = task;    
	task->msg = 0;  
	task->blocked_on.task = NOTHING;        
	task->signal = signal;

	*--(task->stack) = data;
	*--(task->stack) = (int) rt_thread;
	*--(task->stack) = 0;
	*--(task->stack) = (int) rt_startup;

	flags = rt_global_save_flags_and_cli();
	cpuid = run_on_cpu;
	save_cr0_and_clts(linux_cr0);
	save_fpenv(fpu_task->fpu_reg);
	save_fpenv(task->fpu_reg);
	restore_fpenv(fpu_task->fpu_reg);
	restore_cr0(linux_cr0);
	task->next = 0;
	rt_linux_task.prev->next = task;
	task->prev = rt_linux_task.prev;
	rt_linux_task.prev = task;
	rt_global_restore_flags(flags);

	return 0;
}


void rt_set_runnable_on_cpus(RT_TASK *task, unsigned int run_on_cpu)
{
	unsigned long flags;
	RT_TASK *linux_task;

	if (run_on_cpu >= NR_RT_CPUS) {
		run_on_cpu = 0;
	}

	flags = rt_global_save_flags_and_cli();
	switch (rt_smp_oneshot_timer[task->runnable_on_cpus] | 
		(rt_smp_oneshot_timer[run_on_cpu] << 1)) {	
		case 1: 
			task->period = llimd(llimd(task->period, 1E9, tuned.cpu_freq), TIMER_FREQ, 1E9);
			task->resume_time = llimd(llimd(task->resume_time, 1E9, tuned.cpu_freq), TIMER_FREQ, 1E9);
			break;
		case 2:
			task->period = llimd(llimd(task->period, 1E9, TIMER_FREQ), tuned.cpu_freq, 1E9);
			task->resume_time = llimd(llimd(task->resume_time, 1E9, TIMER_FREQ), tuned.cpu_freq, 1E9);
			break;
	}
	if (!((task->prev)->next = task->next)) {
		rt_smp_linux_task[task->runnable_on_cpus].prev = task->prev;
	} else {
		(task->next)->prev = task->prev;
	}
	task->runnable_on_cpus = run_on_cpu;
	task->next = 0;
	(linux_task = rt_smp_linux_task + run_on_cpu)->prev->next = task;
	task->prev = linux_task->prev;
	linux_task->prev = task;
	rt_global_restore_flags(flags);
}


static void rt_schedule(void)
{
	DECLARE_RT_CURRENT;
	RT_TASK *task, *new_task;
	int prio, delay, preemptor;

	prio = RT_LINUX_PRIORITY;
	ASSIGN_RT_CURRENT;
	task = new_task = &rt_linux_task;
	sched_get_global_lock(cpuid);
	if (oneshot_timer) {
		rt_time_h = rdtsc() + rt_half_tick;
		while ((task = task->next)) {
			if ((task->state & DELAYED) && task->resume_time <= rt_time_h) {
				if (task->state & (SEMAPHORE | SEND | RPC)) {
					(task->queue.prev)->next = task->queue.next;
					(task->queue.next)->prev = task->queue.prev;
				}
				task->state &= ~(DELAYED | SEMAPHORE | RECEIVE |
						 SEND    | RPC       | RETURN);
			}
			if (task->state == READY && task->priority < prio) { 
				new_task = task;
				prio = task->priority;
			}
		}
		preemptor = 0;
		task = &rt_linux_task;
		while ((task = task->next)) {
			if ((task->state & DELAYED) && task->priority <= prio && task->resume_time < rt_times.intr_time) {
				rt_times.intr_time = task->resume_time;
				preemptor = 1;
			}
		}
		if (preemptor || (not_fired && prio == RT_LINUX_PRIORITY)) {
			not_fired = 0;
			delay = (int)(rt_times.intr_time - rdtsc()) - tuned.latency;
			if (delay >= tuned.setup_time_TIMER_CPUNIT) {
				delay = imuldiv(delay, TIMER_FREQ, tuned.cpu_freq);
			} else {
				delay = tuned.setup_time_TIMER_UNIT;
				rt_times.intr_time = rdtsc() + (tuned.setup_time_TIMER_CPUNIT);
			}
			apic_read(APIC_TMICT); 
			apic_write(APIC_TMICT, delay);
		}
	} else {
		while ((task = task->next)) {
			if (task->state == READY && task->priority < prio) { 
				new_task = task;
				prio = task->priority;
			}
		}
	}
	sched_release_global_lock(cpuid);

	if (new_task != rt_current) {
		if (rt_current == &rt_linux_task) {
			rt_switch_to_real_time(cpuid);
			save_cr0_and_clts(linux_cr0);
		}
		if (new_task->uses_fpu) {
			if (new_task != fpu_task) {
				save_fpenv(fpu_task->fpu_reg);
				fpu_task = new_task;
				restore_fpenv(fpu_task->fpu_reg);
			}
		}
		if (new_task == &rt_linux_task) {
			rt_switch_to_linux(cpuid);
			restore_cr0(linux_cr0);
		}
		rt_exchange_tasks(rt_smp_current[cpuid], new_task);
		if (rt_current->signal) {
			(*rt_current->signal)();
		}
	}
}


int rt_task_delete(RT_TASK *task)
{
	unsigned long flags;
	QUEUE *q;

	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}

	flags = rt_global_save_flags_and_cli();
	if (task->state & SEMAPHORE) {
		(task->blocked_on.sem)->count++;
		(task->queue.prev)->next = task->queue.next;
		(task->queue.next)->prev = task->queue.prev;
	} else if (task->state & (SEND | RPC)) {
		(task->queue.prev)->next = task->queue.next;
		(task->queue.next)->prev = task->queue.prev;
	} else {
		q = &(task->msg_queue);
		while ((q = q->next) != &(task->msg_queue)) {
			(q->task)->state &= ~(SEND | RPC | DELAYED);
		}       
	}
	task->magic = 0;
	if (!((task->prev)->next = task->next)) {
		rt_smp_linux_task[task->runnable_on_cpus].prev = task->prev;
	} else {
		(task->next)->prev = task->prev;
	}
	if (rt_smp_fpu_task[task->runnable_on_cpus] == task) {
		rt_smp_fpu_task[task->runnable_on_cpus] = rt_smp_linux_task + task->runnable_on_cpus;;
	}
	kfree(task->stack_bottom);
	if (task == RT_CURRENT) {
		rt_schedule();
	}
	rt_global_restore_flags(flags);
	return 0;
}


int rt_get_task_state(RT_TASK *task)
{
	return task->state;
}


static void rt_timer_handler(void)
{
	DECLARE_RT_CURRENT;
	RT_TASK *task, *new_task;
	int prio, delay, preemptor; 

	ASSIGN_RT_CURRENT;
	task = new_task = &rt_linux_task;
	prio = RT_LINUX_PRIORITY;

	rt_times.tick_time = rt_times.intr_time;
	rt_time_h = rt_times.tick_time + rt_half_tick;
	if (rt_times.tick_time >= rt_times.linux_time) {
		rt_times.linux_time += rt_times.linux_tick;
	}

	sched_get_global_lock(cpuid);
	while ((task = task->next)) {
		if ((task->state & DELAYED) && task->resume_time <= rt_time_h) {
			if (task->state & (SEMAPHORE | SEND | RPC)) {
				(task->queue.prev)->next = task->queue.next;
				(task->queue.next)->prev = task->queue.prev;
			}
			task->state &= ~(DELAYED | SEMAPHORE | RECEIVE |
					 SEND    | RPC       | RETURN);
		}
		if (task->state == READY && task->priority < prio) { 
			new_task = task;
			prio = task->priority;
		}
	}

	if (oneshot_timer) {
		rt_times.intr_time = rt_times.linux_time > rt_times.tick_time ?
		rt_times.linux_time : rt_times.tick_time + (rt_times.linux_tick>>1);
		preemptor = preempt_always;
		task = &rt_linux_task;
		while ((task = task->next)) {
			if ((task->state & DELAYED) && task->priority <= prio &&
				      task->resume_time < rt_times.intr_time) {
				rt_times.intr_time = task->resume_time;
				preemptor = 1;
			}
		}
		if (preemptor || prio == RT_LINUX_PRIORITY) {
			not_fired = 0;
			delay = (int)(rt_times.intr_time - rdtsc()) - tuned.latency;
			if (delay >= tuned.setup_time_TIMER_CPUNIT) {
				delay = imuldiv(delay, TIMER_FREQ, tuned.cpu_freq);
			} else {
				delay = tuned.setup_time_TIMER_UNIT;
				rt_times.intr_time = rdtsc() + (tuned.setup_time_TIMER_CPUNIT);
			}
			apic_read(APIC_TMICT); 
			apic_write(APIC_TMICT, delay);
		} else {
			not_fired = 1;
		}
	} else {
		rt_times.intr_time += rt_times.periodic_tick;
	}
	sched_release_global_lock(cpuid);

	if (new_task != rt_current) {
		if (rt_current == &rt_linux_task) {
			rt_switch_to_real_time(cpuid);
			save_cr0_and_clts(linux_cr0);
		}
		if (new_task->uses_fpu) {
			if (new_task != fpu_task) {
				save_fpenv(fpu_task->fpu_reg);
				fpu_task = new_task;
				restore_fpenv(fpu_task->fpu_reg);
			}
		}
		rt_exchange_tasks(rt_smp_current[cpuid], new_task);
		if (rt_current->signal) {
			(*rt_current->signal)();
		}
	}
}


static void recover_jiffies(int irq, void *dev_id, struct pt_regs *regs)
{
	rt_global_cli();
	if (linux_times->tick_time >= linux_times->linux_time) {
		linux_times->linux_time += linux_times->linux_tick;
		rt_pend_linux_irq(TIMER_8254_IRQ);
	}
	rt_global_sti();
} 


void start_rt_timer(struct apic_timer_setup_data *setup_data, unsigned int rcvr_jif)
{
	unsigned long flags, cpuid;

	flags = rt_global_save_flags_and_cli();
	rt_request_apic_timers(rt_timer_handler, setup_data);
	for (cpuid = 0; cpuid < NR_RT_CPUS; cpuid++) {
		if (setup_data[cpuid].mode > 0) {
			oneshot_timer = 0;
			rt_half_tick = (rt_times.periodic_tick + 1)>>1;
		} else {
			oneshot_timer = 1;
			rt_half_tick = (tuned.latency + 1)>>1;
		}
		rt_time_h = rt_times.tick_time + rt_half_tick;
		not_fired = 0;
	}
	linux_times = rt_smp_times + (rcvr_jif < NR_RT_CPUS ? rcvr_jif : 0);
	rt_request_linux_irq(TIMER_8254_IRQ, recover_jiffies, "rtai_sched", recover_jiffies);
	rt_request_cpu_own_irq(SCHED_IPI, rt_schedule);
	rt_global_restore_flags(flags);
}


void stop_rt_timer(void)
{
	unsigned long flags;
	flags = rt_global_save_flags_and_cli();
	rt_free_linux_irq(TIMER_8254_IRQ, recover_jiffies);
	rt_free_apic_timers();
	rt_global_restore_flags(flags);
	rt_busy_sleep(10000);
	rt_free_cpu_own_irq(SCHED_IPI);
}


RT_TASK *rt_whoami(void)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	rt_global_restore_flags(flags);
	return rt_current;
}


int rt_task_signal_handler(RT_TASK *task, void (*handler)(void))
{
	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}
	task->signal = handler;
	return 0;
}


int rt_task_use_fpu(RT_TASK *task, int use_fpu_flag)
{
	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}
	task->uses_fpu = use_fpu_flag != 0 ? 1 : 0;
	return 0;
}


void rt_linux_use_fpu(int use_fpu_flag)
{
	int cpuid;
	for (cpuid = 0; cpuid < NR_RT_CPUS; cpuid++) {
		rt_linux_task.uses_fpu = use_fpu_flag != 0 ? 1 : 0;
	}
}


void rt_preempt_always(int yes_no, int cpuid)
{
	rt_smp_preempt_always[cpuid] = yes_no != 0 ? 1 : 0;
}


const int LinuxFpu = LINUX_FPU;
MODULE_PARM(LinuxFpu, "i");

const int CpuFreq = CPU_FREQ;
MODULE_PARM(CpuFreq, "i");

const int Latency = TIMER_LATENCY;
MODULE_PARM(Latency, "i");

const int SetupTimeTIMER = TIMER_SETUP_TIME;
MODULE_PARM(SetupTimeTIMER, "i");

int init_module(void)
{
	int cpuid;
	for (cpuid = 0; cpuid < NR_RT_CPUS; cpuid++) {
		rt_linux_task.uses_fpu = LinuxFpu != 0 ? 1 : 0;
		rt_linux_task.magic = 0;
		rt_linux_task.runnable_on_cpus = cpuid;
		rt_linux_task.state = (READY | RUNNING);
		rt_linux_task.priority = RT_LINUX_PRIORITY;
		rt_linux_task.base_priority = RT_LINUX_PRIORITY;
		rt_linux_task.signal = 0;
		rt_linux_task.prev = &rt_linux_task;
		rt_linux_task.next = 0;
		rt_smp_current[cpuid] = &rt_linux_task;
		rt_smp_fpu_task[cpuid] = &rt_linux_task;
		preempt_always = 0;
	}
	tuned.cpu_freq = CpuFreq;
	tuned.latency = imuldiv(Latency, CpuFreq, 1E9);
	tuned.setup_time_TIMER_CPUNIT = imuldiv(SetupTimeTIMER, CpuFreq, 1E9);
	tuned.setup_time_TIMER_UNIT = imuldiv(SetupTimeTIMER, TIMER_FREQ, 1E9);
	printk("\n***** STARTING THE REAL TIME MUP SCHEDULER WITH %sLINUX FP SUPPORT *****", rt_linux_task.uses_fpu ? "": "NO ");
	printk("\n***<> LINUX TICK AT %d (HZ) <>***", HZ);
	printk("\n***<> CALIBRATED CPU FREQUENCY %d (HZ) <>***", tuned.cpu_freq);
	printk("\n***<> CALIBRATED APIC_INTERRUPT-TO-SCHEDULER LATENCY %d (ns) <>***", imuldiv(tuned.latency - tuned.setup_time_TIMER_CPUNIT, 1E9, tuned.cpu_freq));
	printk("\n***<> CALIBRATED ONE SHOT APIC SETUP TIME %d (ns) <>***\n\n", imuldiv(tuned.setup_time_TIMER_CPUNIT, 1E9, tuned.cpu_freq));
	rt_mount_rtai();
	rt_request_cpu_own_irq(SCHED_IPI, rt_schedule);
	return 0;
}


void cleanup_module(void)
{
	int cpuid;
	stop_rt_timer();
	for (cpuid = 0; cpuid < NR_RT_CPUS; cpuid++) {
		while (rt_linux_task.next) {
			rt_task_delete(rt_linux_task.next);
		}
	}
	printk("\n***** THE REAL TIME MUP SCHEDULER HAS BEEN REMOVED *****\n\n");
	rt_free_cpu_own_irq(SCHED_IPI);
	rt_umount_rtai();
}

/* ++++++++++++++++++++++++++ TIME CONVERSIONS +++++++++++++++++++++++++++++ */

RTIME count2nano(RTIME counts)
{
	int sign;

	if (counts > 0) {
		sign = 1;
	} else {
		sign = 0;
		counts = - counts;
	}
	counts = oneshot_timer_cpuid ?
		 llimd(counts, 1E9, tuned.cpu_freq):
		 llimd(counts, 1E9, TIMER_FREQ);
	return sign ? counts : - counts;
}


RTIME nano2count(RTIME ns)
{
	int sign;

	if (ns > 0) {
		sign = 1;
	} else {
		sign = 0;
		ns = - ns;
	}
	ns =  oneshot_timer_cpuid ?
	      llimd(ns, tuned.cpu_freq, 1E9) :
	      llimd(ns, TIMER_FREQ, 1E9);
	return sign ? ns : - ns;
}

RTIME count2nano_cpuid(RTIME counts, int cpuid)
{
	int sign;

	if (counts > 0) {
		sign = 1;
	} else {
		sign = 0;
		counts = - counts;
	}
	counts = oneshot_timer ?
		 llimd(counts, 1E9, tuned.cpu_freq):
		 llimd(counts, 1E9, TIMER_FREQ);
	return sign ? counts : - counts;
}


RTIME nano2count_cpuid(RTIME ns, int cpuid)
{
	int sign;

	if (ns > 0) {
		sign = 1;
	} else {
		sign = 0;
		ns = - ns;
	}
	ns =  oneshot_timer ?
	      llimd(ns, tuned.cpu_freq, 1E9) :
	      llimd(ns, TIMER_FREQ, 1E9);
	return sign ? ns : - ns;
}

/* +++++++++++++++++++++++++++++++ TIMINGS ++++++++++++++++++++++++++++++++++ */

static inline RTIME get_time(void)
{
	int cpuid = hard_cpu_id();
	return oneshot_timer ? rdtsc(): rt_times.tick_time;
}


RTIME rt_get_time(void)
{
	return get_time();
}


RTIME rt_get_time_ns(void)
{
	int cpuid = hard_cpu_id();
	return oneshot_timer ?
	       llimd(rdtsc(), 1E9, tuned.cpu_freq) :
	       llimd(rt_times.tick_time, 1E9, TIMER_FREQ);
}


static inline RTIME get_time_cpuid(int cpuid)
{
	return oneshot_timer ? rdtsc(): rt_times.tick_time;
}


RTIME rt_get_time_cpuid(int cpuid)
{
	return get_time_cpuid(cpuid);
}


RTIME rt_get_time_ns_cpuid(int cpuid)
{
	return oneshot_timer ?
	       llimd(rdtsc(), 1E9, tuned.cpu_freq) :
	       llimd(rt_times.tick_time, 1E9, TIMER_FREQ);
}


RTIME rt_get_cpu_time_ns(void)
{
	return llimd(rdtsc(), 1E9, tuned.cpu_freq);
}


void rt_task_yield(void)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if (rt_current->next) {
		(rt_current->prev)->next = rt_current->next;
		(rt_current->next)->prev = rt_current->prev;
		rt_linux_task.prev->next = rt_current;
		rt_current->prev = rt_linux_task.prev;
		rt_current->next = 0;
		rt_linux_task.prev = rt_current;
		rt_schedule();
	}
	rt_global_restore_flags(flags);
}


int rt_task_suspend(RT_TASK *task)
{
	unsigned long flags;

	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}

	flags = rt_global_save_flags_and_cli();
	task->state |= SUSPENDED;
	if (task == RT_CURRENT) {
		rt_schedule();
	}
	rt_global_restore_flags(flags);
	return 0;
}


int rt_task_resume(RT_TASK *task)
{
	unsigned long flags;

	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}

	flags = rt_global_save_flags_and_cli();
	if (((task->state &= ~SUSPENDED) & ~DELAYED) == READY) {
		RT_SCHEDULE(task, hard_cpu_id());
	}
	rt_global_restore_flags(flags);
	return 0;
}


int rt_task_make_periodic(RT_TASK *task, RTIME start_time, RTIME period)
{
	long flags;

	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}
	start_time = nano2count_cpuid(start_time, task->runnable_on_cpus);
	period = nano2count_cpuid(period, task->runnable_on_cpus);
	flags = rt_global_save_flags_and_cli();
	task->resume_time = start_time;
	task->period = period;
	task->state = (task->state & ~SUSPENDED) | DELAYED;
	RT_SCHEDULE(task, hard_cpu_id());
	rt_global_restore_flags(flags);
	return 0;
}


void rt_task_wait_period(void)
{
	DECLARE_RT_CURRENT;
	long flags;

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((rt_current->resume_time += rt_current->period) > rt_time_h) {
		rt_current->state |= DELAYED;
		rt_schedule();
	}
	rt_global_restore_flags(flags);
}


RTIME next_period(void)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	rt_global_restore_flags(flags);
	return rt_current->resume_time + rt_current->period;
}


void rt_busy_sleep(int ns)
{
	RTIME end_time;
	end_time = rdtsc() + llimd(ns, tuned.cpu_freq, 1E9);
	while (rdtsc() < end_time);
}


void rt_sleep(RTIME delay)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((rt_current->resume_time = (oneshot_timer ? rdtsc(): rt_times.tick_time) + delay) > rt_time_h) {
		rt_current->state |= DELAYED;
		rt_schedule();
	}
	rt_global_restore_flags(flags);
}


void rt_sleep_until(RTIME time)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((rt_current->resume_time = time) > rt_time_h) {
		rt_current->state |= DELAYED;
		rt_schedule();
	}
	rt_global_restore_flags(flags);
}

/* +++++++++++++++++++++++++++++ SEMAPHORES ++++++++++++++++++++++++++++++++ */

void rt_sem_init(SEM *sem, int value)
{
	sem->count = value;
	sem->magic = RT_SEM_MAGIC;
	sem->queue.prev = &(sem->queue);
	sem->queue.next = &(sem->queue);
	sem->queue.task = 0;
}


int rt_sem_delete(SEM *sem)
{
	unsigned long flags;
	QUEUE *q;

	if (sem->magic != RT_SEM_MAGIC) {
		return SEM_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	sem->magic = 0;
	q = &(sem->queue);
	while ((q = q->next) != &(sem->queue)) {
		(q->task)->state &= ~(SEMAPHORE | DELAYED);
	}
	rt_global_restore_flags(flags);
	return 0;
}


int rt_sem_signal(SEM *sem)
{
	unsigned long flags;
	RT_TASK *task;

	if (sem->magic != RT_SEM_MAGIC) {
		return SEM_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	if ((sem->count)++ < 0) {
		if ((task = (sem->queue.next)->task)) {
			sem->queue.next = task->queue.next;
			(task->queue.next)->prev = &(sem->queue);
			task->blocked_on.sem = NOTHING;
			if ((task->state &= ~(SEMAPHORE | DELAYED)) == READY) {
				RT_SCHEDULE(task, hard_cpu_id());
			}
		}
	}
	rt_global_restore_flags(flags);
	return 0;
}


int rt_sem_wait(SEM *sem)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	int count;
	QUEUE *q;

	if (sem->magic != RT_SEM_MAGIC) {
		return SEM_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	count = sem->count;
	if (--(sem->count) < 0) {
		rt_current->state |= SEMAPHORE;
		rt_current->blocked_on.sem = sem;
		q = &(sem->queue);
#ifdef SEM_PRIORD
		while ((q = q->next) != &(sem->queue) &&
			 (q->task)->priority <= rt_current->priority);
#endif
		rt_current->queue.prev = q->prev;
		rt_current->queue.next = q;
		(q->prev)->next = &(rt_current->queue);
		q->prev = &(rt_current->queue);
		rt_schedule();
		if (rt_current->blocked_on.sem || sem->magic != RT_SEM_MAGIC) {
			count = SEM_ERR;
		} else { 
			count = sem->count;
		}
	}
	rt_global_restore_flags(flags);
	return count;
}


int rt_sem_wait_if(SEM *sem)
{
	int count;
	unsigned long flags;

	if (sem->magic != RT_SEM_MAGIC) {
		return SEM_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	if ((count = sem->count) > 0) {
		sem->count--;
	}
	rt_global_restore_flags(flags);
	return count;
}


int rt_sem_wait_until(SEM *sem, RTIME time)
{
	DECLARE_RT_CURRENT;
	int count;
	unsigned long flags;
	QUEUE *q;

	if (sem->magic != RT_SEM_MAGIC) {
		return SEM_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	count = sem->count;
	if (--(sem->count) < 0) {
		rt_current->blocked_on.sem = sem;
		if ((rt_current->resume_time = time) > rt_time_h) {
			rt_current->state |= (SEMAPHORE | DELAYED);
			q = &(sem->queue);
#ifdef SEM_PRIORD
			while ((q = q->next) != &(sem->queue) &&
				 (q->task)->priority <= rt_current->priority);
#endif
			rt_current->queue.prev = q->prev;
			rt_current->queue.next = q;
			(q->prev)->next = &(rt_current->queue);
			q->prev = &(rt_current->queue);
			rt_schedule();
		}
		if (sem->magic != RT_SEM_MAGIC) {
			count = SEM_ERR;
		} else {
			if (rt_current->blocked_on.sem) {
				rt_current->blocked_on.sem = NOTHING;
				count = sem->count;
				sem->count++;
			} else {
				count = sem->count;
			}
		}
	}
	rt_global_restore_flags(flags);
	return count;
}


int rt_sem_wait_timed(SEM *sem, RTIME delay)
{
	return rt_sem_wait_until(sem, get_time() + delay);
}

/* ++++++++++++++++++++++++++++++ MESSAGES +++++++++++++++++++++++++++++++++ */

/* ++++++++++++++++++++++++++++++++ SEND +++++++++++++++++++++++++++++++++++ */

RT_TASK *rt_send(RT_TASK *task, unsigned int msg)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	QUEUE *q;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((task->state & RECEIVE) &&
	      (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = msg;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		if ((task->state &= ~(RECEIVE | DELAYED)) == READY) {
			RT_SCHEDULE(task, cpuid);
		}
	} else {
		rt_current->msg = msg;
		rt_current->msg_queue.task = task;
		q = &(task->msg_queue);
#ifdef MSG_PRIORD
		while ((q = q->next) != &(task->msg_queue) &&
			 (q->task)->priority <= rt_current->priority);
		if (task->priority > rt_current->priority) {
			task->priority = rt_current->priority;
		}
#endif
		rt_current->queue.prev = q->prev;
		rt_current->queue.next = q;
		(q->prev)->next = &(rt_current->queue);
		q->prev = &(rt_current->queue);
		rt_current->state |= SEND;
		rt_schedule();
	}
	if (rt_current->msg_queue.task != rt_current) {
		rt_current->msg_queue.task = rt_current;
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_send_if(RT_TASK *task, unsigned int msg)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((task->state & RECEIVE) &&
	      (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = msg;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		if ((task->state &= ~(RECEIVE | DELAYED)) == READY) {
			RT_SCHEDULE(task, cpuid);
		}
		if (rt_current->msg_queue.task != rt_current) {
			rt_current->msg_queue.task = rt_current;
			task = (RT_TASK *)0;
		}
	} else {
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_send_until(RT_TASK *task, unsigned int msg, RTIME time)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	QUEUE *q;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((task->state & RECEIVE) &&
	      (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = msg;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		if ((task->state &= ~(RECEIVE | DELAYED)) == READY) {
			RT_SCHEDULE(task, cpuid);
		}
	} else {
		rt_current->msg_queue.task = task;
		if ((rt_current->resume_time = time) > rt_time_h) {
			rt_current->msg = msg;
			q = &(task->msg_queue);
#ifdef MSG_PRIORD
			while ((q = q->next) != &(task->msg_queue) &&
				 (q->task)->priority <= rt_current->priority);
			if (task->priority > rt_current->priority) {
				task->priority = rt_current->priority;
			}
#endif
			rt_current->queue.prev = q->prev;
			rt_current->queue.next = q;
			(q->prev)->next = &(rt_current->queue);
			q->prev = &(rt_current->queue);
			rt_current->state |= (SEND | DELAYED);
			rt_schedule();
		}
	}
	if (rt_current->msg_queue.task != rt_current) {
		rt_current->msg_queue.task = rt_current;
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_send_timed(RT_TASK *task, unsigned int msg, RTIME delay)
{
	return rt_send_until(task, msg, get_time() + delay);
}

/* +++++++++++++++++++++++++++++++++ RPC +++++++++++++++++++++++++++++++++++ */

RT_TASK *rt_rpc(RT_TASK *task, unsigned int to_do, unsigned int *result)
{

	DECLARE_RT_CURRENT;
	unsigned long flags;
	QUEUE *q;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((task->state & RECEIVE) &&
		(!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = to_do;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		task->state &= ~(RECEIVE | DELAYED);
		rt_current->state |= RETURN;
	} else {
		rt_current->msg = to_do;
		q = &(task->msg_queue);
#ifdef MSG_PRIORD
		while ((q = q->next) != &(task->msg_queue) &&
			 (q->task)->priority <= rt_current->priority);
		if (task->priority > rt_current->priority) {
			task->priority = rt_current->priority;
		}
#endif
		rt_current->queue.prev = q->prev;
		rt_current->queue.next = q;
		(q->prev)->next = &(rt_current->queue);
		q->prev = &(rt_current->queue);
		rt_current->state |= RPC;
	}
	rt_current->msg_queue.task = task;
	RT_SCHEDULE_BOTH(task, cpuid);
	if (rt_current->msg_queue.task == rt_current) {
		*result = rt_current->msg;
	} else {
		rt_current->msg_queue.task = rt_current;
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_rpc_if(RT_TASK *task, unsigned int to_do, unsigned int *result)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}
	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((task->state & RECEIVE) &&
	      (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = to_do;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		task->state &= ~(RECEIVE | DELAYED);
		rt_current->state |= RETURN;
		rt_current->msg_queue.task = task;
		RT_SCHEDULE_BOTH(task, cpuid);
		if (rt_current->msg_queue.task == rt_current) {
			*result = rt_current->msg;
		} else {
			rt_current->msg_queue.task = rt_current;
			task = (RT_TASK *)0;
		}
	} else {
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_rpc_until(RT_TASK *task, unsigned int to_do, unsigned int *result, RTIME time)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	QUEUE *q;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((task->state & RECEIVE) &&
	    (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = to_do;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		task->state &= ~(RECEIVE | DELAYED);
		rt_current->state |= (RETURN | DELAYED);
	} else {
		if ((rt_current->resume_time = time) <= rt_time_h) {
			rt_global_restore_flags(flags);
			return (RT_TASK *)0;
		}
		rt_current->msg = to_do;
		q = &(task->msg_queue);
#ifdef MSG_PRIORD
		while ((q = q->next) != &(task->msg_queue) &&
			 (q->task)->priority <= rt_current->priority);
		if (task->priority > rt_current->priority) {
			task->priority = rt_current->priority;
		}
#endif
		rt_current->queue.prev = q->prev;
		rt_current->queue.next = q;
		(q->prev)->next = &(rt_current->queue);
		q->prev = &(rt_current->queue);
		rt_current->state |= (RPC | DELAYED);
	}
	rt_current->msg_queue.task = task;
	RT_SCHEDULE_BOTH(task, cpuid);
	if (rt_current->msg_queue.task == rt_current) {
		*result = rt_current->msg;
	} else {
		rt_current->msg_queue.task = rt_current;
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_rpc_timed(RT_TASK *task, unsigned int to_do, unsigned int *result, RTIME delay)
{
	return rt_rpc_until(task, to_do, result, get_time() + delay);
}

/* ++++++++++++++++++++++++++++++ RPC_RETURN +++++++++++++++++++++++++++++++ */

int rt_isrpc(RT_TASK *task)
{
	return task->state & RETURN;
}


RT_TASK *rt_return(RT_TASK *task, unsigned int result)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((task->state & RETURN) && task->msg_queue.task == rt_current) {
		task->msg = result;
		task->msg_queue.task = task;
		if ((task->state &= ~(RETURN | DELAYED)) == READY) {
			RT_SCHEDULE(task, cpuid);
		}
	} else {
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}

/* +++++++++++++++++++++++++++++++ RECEIVE +++++++++++++++++++++++++++++++++ */

RT_TASK *rt_receive(RT_TASK *task, unsigned int *msg)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;

	if (task && task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if (!task) task = (rt_current->msg_queue.next)->task;
	if ((task->state & (SEND | RPC)) && task->msg_queue.task == rt_current) {
		(task->queue.next)->prev = task->queue.prev;
		(task->queue.prev)->next = task->queue.next;
		*msg = task->msg;
#ifdef MSG_PRIORD
		rt_current->priority = rt_current->base_priority > task->priority ?
			task->priority : rt_current->base_priority;
#endif
		rt_current->msg_queue.task = task;
		if (task->state & SEND) {
			task->msg_queue.task = task;
			if ((task->state  &=  ~(SEND | DELAYED)) == READY) {
				RT_SCHEDULE(task, cpuid);
			}
		} else if (task->state & RPC) {
			task->state = (task->state & ~(RPC | DELAYED)) | RETURN;
		}
	} else {
#ifdef MSG_PRIORD
		task->priority = rt_current->base_priority;
#endif
		rt_current->blocked_on.task = SOMETHING;
		rt_current->state |= RECEIVE;
		rt_current->msg_queue.task = (task != rt_current) ? task : (RT_TASK *)0;
		rt_schedule();
		*msg = rt_current->msg;
	}
	if (rt_current->blocked_on.task) {
		rt_current->blocked_on.task = NOTHING;
		task = (RT_TASK *)0;
	} else {
		task = rt_current->msg_queue.task;
	}
	rt_current->msg_queue.task = rt_current;
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_receive_if(RT_TASK *task, unsigned int *msg)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;

	if (task && task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if (!task) task = (rt_current->msg_queue.next)->task;
	if ((task->state & (SEND | RPC)) && task->msg_queue.task == rt_current) {
		(task->queue.next)->prev = task->queue.prev;
		(task->queue.prev)->next = task->queue.next;
		*msg = task->msg;
#ifdef MSG_PRIORD
		rt_current->priority = rt_current->base_priority > task->priority ?
			task->priority : rt_current->base_priority;
#endif
		rt_current->msg_queue.task = task;
		if (task->state & SEND) {
			task->msg_queue.task = task;
			if ((task->state  &=  ~(SEND | DELAYED)) == READY) {
				RT_SCHEDULE(task, cpuid);
			}
		} else if (task->state & RPC) {
			task->state = (task->state & ~(RPC | DELAYED)) | RETURN;
		}
		if (rt_current->blocked_on.task) {
			rt_current->blocked_on.task = NOTHING;
			task = (RT_TASK *)0;
		} else {
			task = rt_current->msg_queue.task;
		}
		rt_current->msg_queue.task = rt_current;
	} else {
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_receive_until(RT_TASK *task, unsigned int *msg, RTIME time)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;

	if (task && task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if (!task) task = (rt_current->msg_queue.next)->task;
	if ((task->state & (SEND | RPC)) && task->msg_queue.task == rt_current) {
		(task->queue.next)->prev = task->queue.prev;
		(task->queue.prev)->next = task->queue.next;
		*msg = task->msg;
#ifdef MSG_PRIORD
		rt_current->priority = rt_current->base_priority > task->priority ?
			task->priority : rt_current->base_priority;
#endif
		rt_current->msg_queue.task = task;
		if (task->state & SEND) {
			task->msg_queue.task = task;
			if ((task->state  &=  ~(SEND | DELAYED)) == READY) {
				RT_SCHEDULE(task, cpuid);
			}
		} else if (task->state & RPC) {
			task->state = (task->state & ~(RPC | DELAYED)) | RETURN;
		}
	} else {
#ifdef MSG_PRIORD
		task->priority = rt_current->base_priority;
#endif
		rt_current->blocked_on.task = SOMETHING;
		if ((rt_current->resume_time = time) > rt_time_h) {
			rt_current->state |= (RECEIVE | DELAYED);
			rt_current->msg_queue.task = (task != rt_current) ? task : (RT_TASK *)0;
			rt_schedule();
			*msg = rt_current->msg;
		}
	}
	if (rt_current->blocked_on.task) {
		rt_current->blocked_on.task = NOTHING;
		task = (RT_TASK *)0;
	} else {
		task = rt_current->msg_queue.task;
	}
	rt_current->msg_queue.task = rt_current;
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_receive_timed(RT_TASK *task, unsigned int *msg, RTIME delay)
{
	return rt_receive_until(task, msg, get_time() + delay);
}
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
