//////////////////////////////////////////////////////////////////////////////
//
//      Copyright  1999 Zentropic Computing, All rights reserved
//
// Authors:             Steve Papacharalambous
// Original date:       Thu 15 Jul 1999
//
// pthreads interface for Real Time Linux.
//
//
// The contents of this file are subject to the Mozilla Public
// License Version 1.0 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
//
///////////////////////////////////////////////////////////////////////////////

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/time.h>
#include <linux/sched.h>

#include <asm/signal.h>


#include <rtai.h>
#include <rtai_sched.h>

#define NSECS_PER_SEC 1000000000
#define DEFAULT_TICK_PERIOD 20000


static inline struct timespec timespec_from_ns (long long t)
{
	struct timespec t_spec;
	__asm__("idivl %%ecx\n\t"
		:"=a" (t_spec.tv_sec), "=d" (t_spec.tv_nsec)
		: "A" (t), "c" (NSECS_PER_SEC)
		:"cx");
	if (t_spec.tv_nsec < 0) {
		t_spec.tv_nsec += NSECS_PER_SEC;
		t_spec.tv_sec --;
	}
	return t_spec;
}


static inline void ts_from_ns (long long t, struct timespec *t_spec)
{
	__asm__("idivl %%ecx\n\t"
		:"=a" (t_spec->tv_sec), "=d" (t_spec->tv_nsec)
		: "A" (t), "c" (NSECS_PER_SEC)
		:"cx");
	if (t_spec->tv_nsec < 0) {
		t_spec->tv_nsec += NSECS_PER_SEC;
		t_spec->tv_sec --;
	}
}




int init_module(void) {

  struct timespec ts_result, sp_ts;

  start_rt_timer((int)nano2count(DEFAULT_TICK_PERIOD));
  ts_result = timespec_from_ns(count2nano(rt_get_time()));
  ts_from_ns(count2nano(rt_get_time()), &sp_ts);

  printk("\nRT Timer sec: %ld, nsec: %ld\n", ts_result.tv_sec,
                                             ts_result.tv_nsec);

  printk("\nTimer 2 sec: %ld, nsec: %ld\n", sp_ts.tv_sec, sp_ts.tv_nsec);

  return(0);

}


void cleanup_module(void) {

  stop_rt_timer();

}
