/*
"COPYRIGHT (C) 1999  Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USAr.
*/


/* Module to hook plain Linux up to do real time the way you like, hardware, */
/* hopefully, fully trapped; the machine is in your hand, do what you want!  */

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#include <asm/system.h>
#include <asm/i82489.h>
#include <asm/fixmap.h>
#include <asm/smp.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/bitops.h>
#include <asm/atomic.h>
#include <asm/desc.h>

#include "include/rtai.h"
#include "include/rtai_srq.h"


/* Macros to setup intercept-dispatching of any interrupt, Linux cannot hard */
/* interrupt any more and all those related to IPIs and IOs are catched and  */
/* can be our slaves.                                                        */

#define __STR(x) #x
#define STR(x) __STR(x)

// Simple SAVE/RSTR for compilation with the -fomit-frame-pointer option.
// Very simple assembler, support for using the fpu without problem, has been
// taken away for reason of efficiency. To use hard floating point in your 
// interrupt handlers, without any problem, see the fpu support macros 
// available in "rtai.h".

struct fill_t { unsigned long fill[6]; };

#define SAVE_REG(irq) __asm__ __volatile__ (" \
	pushl $"#irq"; pushl %es; pushl %ds; pushl %eax;\n\t \
	pushl %ebp; pushl %ecx; pushl %edx;\n\t \
	movl $" STR(__KERNEL_DS) ",%edx; movl %dx,%ds; movl %dx,%es")

#define RSTR_REG __asm__ __volatile__ (" \
	popl %edx; popl %ecx; testl %eax,%eax; jnz 1f;\n\t \
	popl %ebp; popl %eax; popl %ds; popl %es; addl $4,%esp; iret;\n\t \
	1: cld; pushl %edi; pushl %esi; pushl %edx;\n\t \
	pushl %ecx; pushl %ebx; jmp ret_from_intr")

#define GLOBAL_IRQ_FNAME(y) y##_interrupt(void)
#define GLOBAL_IRQ_NAME(y) GLOBAL_IRQ_FNAME(GLOBAL##y)

#define BUILD_GLOBAL_IRQ(irq) static void GLOBAL_IRQ_NAME(irq) \
{ \
	SAVE_REG(irq); \
	__asm__ __volatile__ ("call "SYMBOL_NAME_STR(dispatch_global_irq)); \
	RSTR_REG; \
}

#define CPU_OWN_IRQ_FNAME(y) y##_interrupt(void)
#define CPU_OWN_IRQ_NAME(y) CPU_OWN_IRQ_FNAME(CPU_OWN##y)

#define BUILD_CPU_OWN_IRQ(irq) static void CPU_OWN_IRQ_NAME(irq) \
{ \
	SAVE_REG(irq); \
	__asm__ __volatile__ ("call "SYMBOL_NAME_STR(dispatch_cpu_own_irq)); \
	RSTR_REG; \
}

static void srqisr(void)
{
	__asm__ __volatile__ (" \
	cld; pushl %es; pushl %ds; pushl %ebp;\n\t \
	pushl %edi; pushl %esi; pushl %ecx;\n\t \
	pushl %ebx; pushl %edx; pushl %eax;\n\t \
	movl $" STR(__KERNEL_DS) ",%ebx; movl %bx,%ds; movl %bx,%es");

	__asm__ __volatile__ ("call "SYMBOL_NAME_STR(dispatch_srq));

	__asm__ __volatile__ (" \
	addl $8,%esp; popl %ebx; popl %ecx; popl %esi;\n\t \
	popl %edi; popl %ebp; popl %ds; popl %es; iret");
}

/* Setup intercept-dispatching handlers */

BUILD_GLOBAL_IRQ(0)   BUILD_GLOBAL_IRQ(1)   BUILD_GLOBAL_IRQ(2)
BUILD_GLOBAL_IRQ(3)   BUILD_GLOBAL_IRQ(4)   BUILD_GLOBAL_IRQ(5)  
BUILD_GLOBAL_IRQ(6)   BUILD_GLOBAL_IRQ(7)   BUILD_GLOBAL_IRQ(8)
BUILD_GLOBAL_IRQ(9)   BUILD_GLOBAL_IRQ(10)  BUILD_GLOBAL_IRQ(11)
BUILD_GLOBAL_IRQ(12)  BUILD_GLOBAL_IRQ(13)  BUILD_GLOBAL_IRQ(14) 
BUILD_GLOBAL_IRQ(15)  BUILD_GLOBAL_IRQ(16)  BUILD_GLOBAL_IRQ(17) 
BUILD_GLOBAL_IRQ(18)  BUILD_GLOBAL_IRQ(19)  BUILD_GLOBAL_IRQ(20) 
BUILD_GLOBAL_IRQ(21)  BUILD_GLOBAL_IRQ(22)  BUILD_GLOBAL_IRQ(23)
BUILD_GLOBAL_IRQ(24)  BUILD_GLOBAL_IRQ(25)  BUILD_GLOBAL_IRQ(26) 
BUILD_GLOBAL_IRQ(27)  BUILD_GLOBAL_IRQ(28)  BUILD_GLOBAL_IRQ(29) 
BUILD_GLOBAL_IRQ(30)  BUILD_GLOBAL_IRQ(31)

BUILD_CPU_OWN_IRQ(0)  BUILD_CPU_OWN_IRQ(1)  BUILD_CPU_OWN_IRQ(2)
BUILD_CPU_OWN_IRQ(3)  BUILD_CPU_OWN_IRQ(4)  BUILD_CPU_OWN_IRQ(5)
BUILD_CPU_OWN_IRQ(6)  BUILD_CPU_OWN_IRQ(7)  BUILD_CPU_OWN_IRQ(8)
BUILD_CPU_OWN_IRQ(9)

/* Some define */

/* The two below cannot be > 32 (bits) */
#define NR_GLOBAL_IRQS   32
#define NR_CPU_OWN_IRQS  10
#define NR_IC_DEV_TYPES   3 // 0=8259A, 1=level/edge-IO-APIC, 2=none

#define DUMMY_VECTOR  0x32

#define HARD_LOCK_IPI      	RTAI_1_IPI
#define RTAI_APIC_TIMER_IPI    	RTAI_3_IPI
#define RTAI_APIC_TIMER_VECTOR  RTAI_3_VECTOR
#define APIC_ICOUNT		((FREQ_APIC + HZ/2)/HZ)

#define RESCHEDULE_IPI     0
#define INVALIDATE_IPI     1
#define STOP_CPU_IPI       2
#define LOCAL_TIMER_IPI    3
#define CALL_FUNCTION_IPI  4
#define SPURIOUS_IPI       5

/* BE SURE TYPE DEFINITIONS BELOW ARE THE SAME AS THOSE OF Linux, ALWAYS!!!  */
/* AND ALSO MATCH THE ORDER OF THE SAME NUMBER OF ....._IPI MACROS DEFINED   */
/* IMMEDIATLY ABOVE, THE IRQ0_TRAP_VECTOR EXCLUDED.                          */
/* BE SURE TYPE DEFINITIONS BELOW ARE THE SAME AS THOSE OF Linux, ALWAYS!!!  */
/* BE SURE TYPE DEFINITIONS BELOW ARE THE SAME AS THOSE OF Linux, ALWAYS!!!  */

#define RESCHEDULE_VECTOR      0x30
#define INVALIDATE_TLB_VECTOR  0x31
#define STOP_CPU_VECTOR        0x40
#define LOCAL_TIMER_VECTOR     0x41
#define CALL_FUNCTION_VECTOR   0x50
#define SPURIOUS_APIC_VECTOR   0xFF

#define IRQ0_TRAP_VECTOR  0x51

struct hw_interrupt_type {
	const char        *typename;
	void  (*startup)  (unsigned int irq);
	void  (*shutdown) (unsigned int irq);
	void  (*handle)   (unsigned int irq, struct pt_regs *regs);
	void  (*enable)   (unsigned int irq);
	void  (*disable)  (unsigned int irq);
};
typedef struct {
	unsigned int              status;
	struct hw_interrupt_type  *handler;
	struct irqaction          *action;
	unsigned int              depth;
	unsigned int              unused[4];
} irq_desc_t;

#define IRQ_DESC ((irq_desc_t *)rthal.irq_desc)

/* BE SURE TYPE DEFINITIONS ABOVE ARE THE SAME AS THOSE OF Linux, ALWAYS!!! */
/* BE SURE TYPE DEFINITIONS ABOVE ARE THE SAME AS THOSE OF Linux, ALWAYS!!! */
/* BE SURE TYPE DEFINITIONS BELOW ARE THE SAME AS THOSE OF Linux, ALWAYS!!!  */


// RTHAL from Linux
extern struct rt_hal rthal;

/* Most of our data */

static int global_vector[NR_GLOBAL_IRQS] = { 0x20, 0x21, 0x22, 0x23, 0x24,
	 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F, 
DUMMY_VECTOR, DUMMY_VECTOR, DUMMY_VECTOR, DUMMY_VECTOR, DUMMY_VECTOR,
DUMMY_VECTOR, DUMMY_VECTOR, DUMMY_VECTOR, DUMMY_VECTOR, DUMMY_VECTOR,
DUMMY_VECTOR, DUMMY_VECTOR, DUMMY_VECTOR, DUMMY_VECTOR, DUMMY_VECTOR,
DUMMY_VECTOR };

static void (*global_interrupt[NR_GLOBAL_IRQS])(void) = {
	GLOBAL0_interrupt,  GLOBAL1_interrupt,  GLOBAL2_interrupt,
	GLOBAL3_interrupt,  GLOBAL4_interrupt,  GLOBAL5_interrupt, 
	GLOBAL6_interrupt,  GLOBAL7_interrupt,  GLOBAL8_interrupt, 
	GLOBAL9_interrupt,  GLOBAL10_interrupt, GLOBAL11_interrupt,
	GLOBAL12_interrupt, GLOBAL13_interrupt, GLOBAL14_interrupt,
	GLOBAL15_interrupt, GLOBAL16_interrupt, GLOBAL17_interrupt,
	GLOBAL18_interrupt, GLOBAL19_interrupt,	GLOBAL20_interrupt,
	GLOBAL21_interrupt, GLOBAL22_interrupt, GLOBAL23_interrupt,
	GLOBAL24_interrupt, GLOBAL25_interrupt, GLOBAL26_interrupt,
	GLOBAL27_interrupt, GLOBAL28_interrupt, GLOBAL29_interrupt,
	GLOBAL30_interrupt, GLOBAL31_interrupt };

static void (*global_irq_handler[NR_GLOBAL_IRQS])(void);
static struct sysrq_t {
	unsigned int label;
	void (*rtai_handler)(void);
	long long (*user_handler)(unsigned int whatever);
} sysrq[NR_GLOBAL_IRQS];

static int cpu_own_vector[NR_CPU_OWN_IRQS] = {
	RESCHEDULE_VECTOR,  INVALIDATE_TLB_VECTOR, STOP_CPU_VECTOR, 
	LOCAL_TIMER_VECTOR, CALL_FUNCTION_VECTOR,  SPURIOUS_APIC_VECTOR,
	RTAI_1_VECTOR,      RTAI_2_VECTOR,         RTAI_3_VECTOR,
	RTAI_4_VECTOR };

static void (*cpu_own_interrupt[NR_CPU_OWN_IRQS])(void) = {
	CPU_OWN0_interrupt, CPU_OWN1_interrupt, CPU_OWN2_interrupt,
	CPU_OWN3_interrupt, CPU_OWN4_interrupt, CPU_OWN5_interrupt,
	CPU_OWN6_interrupt, CPU_OWN7_interrupt, CPU_OWN8_interrupt,
	CPU_OWN9_interrupt };

static struct cpu_own_irq_handling {
	volatile unsigned long dest_status;		
	void (*handler)(void);
} cpu_own_irq[NR_CPU_OWN_IRQS];

// The main items to be saved-restored to make Linux our humble slave
static struct rt_hal linux_rthal;
static struct desc_struct linux_idt_table[256];
static void (*linux_isr[256])(void);
static struct hw_interrupt_type *linux_irq_desc_handler[NR_GLOBAL_IRQS];

static struct global_rt_status {
	volatile unsigned int pending_irqs;
	volatile unsigned int activ_irqs;
	volatile unsigned int pending_srqs;
	volatile unsigned int activ_srqs;
	volatile unsigned int cpu_in_sti;
	volatile unsigned int used_by_linux;
  	volatile unsigned int locked_cpus;
  	volatile unsigned int hard_nesting;
	volatile unsigned int hard_lock_all_service;
	spinlock_t hard_lock;
	spinlock_t data_lock;
	spinlock_t ic_lock[NR_IC_DEV_TYPES];
} global;

volatile unsigned int *locked_cpus = &global.locked_cpus;

static struct cpu_own_status {
	volatile unsigned int intr_flag;
	volatile unsigned int linux_intr_flag;
	volatile unsigned int pending_irqs;
	volatile unsigned int activ_irqs;
} processor[NR_RT_CPUS]; 


/* Interrupt descriptor table manipulation. No assembler here. These are    */
/* the base for manipulating Linux interrupt handling without even touching */
/* the kernel.                                                              */

struct desc_struct rt_set_full_intr_vect(unsigned int vector, int type, int dpl, void (*handler)(void))
{

// "dpl" is the descriptor privilege level: 0-highest, 3-lowest.
// "type" is the interrupt type: 14 interrupt (cli), 15 trap (no cli).

struct desc_struct idt_element = rthal.idt_table[vector];

	rthal.idt_table[vector].a = (__KERNEL_CS << 16) | 
					((unsigned int)handler & 0x0000FFFF);
	rthal.idt_table[vector].b = ((unsigned int)handler & 0xFFFF0000) | 
					(0x8000 + (dpl << 13) + (type << 8));
	return idt_element;
}

void rt_reset_full_intr_vect(unsigned int vector, struct desc_struct idt_element)
{
	rthal.idt_table[vector] = idt_element;
	return;
}

// Get the interrupt handler proper.
static inline void *get_intr_handler(unsigned int vector)
{
	return (void *)((rthal.idt_table[vector].b & 0xFFFF0000) | 
			(rthal.idt_table[vector].a & 0x0000FFFF));
}

static inline void set_intr_vect(unsigned int vector, void (*handler)(void))
{

/* It should be done as above, in set_full_intr_vect, but we keep what has not
to be changed as it is already. So we have not to mind of DPL, TYPE and 
__KERNEL_CS. Then let's just change only the offset part of the idt table. */

	rthal.idt_table[vector].a = (rthal.idt_table[vector].a & 0xFFFF0000) | 
				           ((unsigned int)handler & 0x0000FFFF);
	rthal.idt_table[vector].b = ((unsigned int)handler & 0xFFFF0000) | 
			               (rthal.idt_table[vector].b & 0x0000FFFF);
}

void *rt_set_intr_handler(unsigned int vector, void (*handler)(void))
{
	void (*saved_handler)(void) = get_intr_handler(vector);
	set_intr_vect(vector, handler);
	return saved_handler;;
}

void rt_reset_intr_handler(unsigned int vector, void (*handler)(void))
{
	set_intr_vect(vector, handler);
	return;
}


/* Our interprocessor messaging */ 

void send_ipi_shorthand(unsigned int shorthand, int irq)
{
	unsigned long flags;
	hard_save_flags(flags);
	hard_cli();
	apic_write(APIC_ICR, (apic_read(APIC_ICR) & ~0xCDFFF) |
	      	(APIC_DEST_DM_FIXED | shorthand | cpu_own_vector[irq]));
	hard_restore_flags(flags);
}

void send_ipi_logical(unsigned long dest, int irq)
{
	unsigned long flags;
	if ((dest &= cpu_present_map)) {
		hard_save_flags(flags);
		hard_cli();
		apic_write(APIC_ICR2, (apic_read(APIC_ICR2) & 0xFFFFFF) | 
								(dest << 24));
		apic_write(APIC_ICR, (apic_read(APIC_ICR) & ~0xCDFFF) |
				(APIC_DEST_LOGICAL | cpu_own_vector[irq]));
		hard_restore_flags(flags);
	}
}

static struct apic_timer_setup_data apic_timer_mode[NR_RT_CPUS];

static inline void setup_periodic_apic(unsigned int count, unsigned int vector)
{
	apic_read(APIC_LVTT);
	apic_write(APIC_LVTT, APIC_LVT_TIMER_PERIODIC | vector);
	apic_read(APIC_TMICT);
	apic_write(APIC_TMICT, count);
}

static inline void setup_oneshot_apic(unsigned int count, unsigned int vector)
{
	apic_read(APIC_LVTT);
	apic_write(APIC_LVTT, vector);
	apic_read(APIC_TMICT);
	apic_write(APIC_TMICT, count);
}

/* A few things dependending on being in a UP-SMP configuration */

#ifdef __SMP__

static inline void ack_APIC_irq(void)
{
	apic_read(APIC_SPIV);
	apic_write(APIC_EOI, 0);
}

static void set_flat_dest_mode(void)
{
	int cpuid;
	set_bit(cpuid = hard_cpu_id(), &cpu_own_irq[HARD_LOCK_IPI].dest_status);
	rt_spin_lock(&(global.hard_lock));
	apic_write(APIC_LDR, apic_read(APIC_LDR)|(0x1000000<<hard_cpu_id()));
/*	apic_write(APIC_DFR, apic_read(APIC_DFR)|0xF0000000); Linux default */
	rt_spin_unlock(&(global.hard_lock));
	clear_bit(cpuid, &cpu_own_irq[HARD_LOCK_IPI].dest_status);
}

static long long apic_timers_sync_time;

static void hard_lock_all_handler(void)
{
	int cpuid;
	struct apic_timer_setup_data *p;

	set_bit(cpuid = hard_cpu_id(), &cpu_own_irq[HARD_LOCK_IPI].dest_status);
	rt_spin_lock(&(global.hard_lock));
	switch (global.hard_lock_all_service) {
		case 1:
			p = apic_timer_mode + cpuid;
			if (p->mode > 0) {
				while (rdtsc() < apic_timers_sync_time);
				setup_periodic_apic(p->count, RTAI_APIC_TIMER_VECTOR);
				break;
			} else if (!p->mode) {
				while (rdtsc() < apic_timers_sync_time);
				setup_oneshot_apic(p->count, RTAI_APIC_TIMER_VECTOR);
				break;
			}
		case 2:
			setup_oneshot_apic(0, RTAI_APIC_TIMER_VECTOR);
			break;
		case 3:
			setup_periodic_apic(APIC_ICOUNT, LOCAL_TIMER_VECTOR);
			break;
	}
	rt_spin_unlock(&(global.hard_lock));
	clear_bit(cpuid, &cpu_own_irq[HARD_LOCK_IPI].dest_status);
}

static inline void check_smp_invalidate(void)
{
	if (test_and_clear_bit(hard_cpu_id(), &smp_invalidate_needed)) {
			local_flush_tlb();
	}
}

#else

#define ack_APIC_irq()

static void set_flat_dest_mode(void) { }

static void hard_lock_all_handler(void) { }

#define check_smp_invalidate()

#endif

static void invalidate_handler(void)
{
	check_smp_invalidate();
}

/* Emulation of Linux interrupt control and interrupt soft delivery.       */

static inline void do_linux_irq(int vector)
{
	__asm__ __volatile__ ("pushf; push %cs");
	linux_isr[vector]();
	return;
}

void rt_do_irq(int vector)
{
	__asm__ __volatile__ ("pushf; push %cs");
	((void (*)(void))get_intr_handler(vector))();
	return;
}

/* The old SFIF and SFREQ of RTLinux plus vanilla kernel services to be used */
/* within real time applications.                                            */

static void linux_cli(void)
{ 
	processor[hard_cpu_id()].intr_flag = 0;
}

static void linux_sti(void)
{
       	unsigned long irq, cpuid, intr_enabled;
       	struct cpu_own_status *cpu;
	cpu = processor + (cpuid = hard_cpu_id());
	cpu->intr_flag = intr_enabled = (1 << IFLAG) | (1 << cpuid);
	while (!test_and_set_bit(cpuid, &global.cpu_in_sti)) {
	       	while (global.pending_irqs || cpu->pending_irqs ||
						       global.pending_srqs) {
#ifdef __SMP__
			hard_cli();
			if ((irq = cpu->pending_irqs & ~(cpu->activ_irqs))) {
				irq = ffnz(irq);
				set_bit(irq, &cpu->activ_irqs);
				clear_bit(irq, &cpu->pending_irqs);
				hard_sti();
				cpu->intr_flag = 0;
		       		do_linux_irq(cpu_own_vector[irq]);
				clear_bit(irq, &cpu->activ_irqs);
				cpu->intr_flag = intr_enabled;
	       		} else {
				hard_sti();
			}
#endif
			rt_spin_lock_irq(&(global.data_lock));
			if ((irq = global.pending_srqs & ~global.activ_srqs)) {
				irq = ffnz(irq);
				set_bit(irq, &global.activ_srqs);
				clear_bit(irq, &global.pending_srqs);
				rt_spin_unlock_irq(&(global.data_lock));
				if (sysrq[irq].rtai_handler) {
					sysrq[irq].rtai_handler();
				}
				clear_bit(irq, &global.activ_srqs);
			} else {
				rt_spin_unlock_irq(&(global.data_lock));
			}
			rt_spin_lock_irq(&(global.data_lock));
			if ((irq = global.pending_irqs & ~global.activ_irqs)) {
				irq = ffnz(irq);
				set_bit(irq, &global.activ_irqs);
				clear_bit(irq, &global.pending_irqs);
				rt_spin_unlock_irq(&(global.data_lock));
				cpu->intr_flag = 0;
				do_linux_irq(global_vector[irq]);
				clear_bit(irq, &global.activ_irqs);
				cpu->intr_flag = intr_enabled;
			} else {
				rt_spin_unlock_irq(&(global.data_lock));
			}
		}
		clear_bit(cpuid, &global.cpu_in_sti);
	       	if (!(global.pending_irqs | cpu->pending_irqs |
	      				    global.pending_srqs)) break;
	}
}

static unsigned int linux_save_flags(void)
{
	return processor[hard_cpu_id()].intr_flag;
}

static void linux_restore_flags(unsigned int flags)
{
	if (flags) {
		linux_sti();
	} else {
		processor[hard_cpu_id()].intr_flag = 0;
	}
}


/* Hard lock-unlock all cpus except the one that sets the lock; to be used */
/* to atomically set up critical things affecting both Linux and RTAI.     */ 

static inline int hard_lock_all(void)
{
	unsigned long flags;
	flags = rt_global_save_flags_and_cli();

	if (!global.hard_nesting++) {
		global.hard_lock_all_service = 0;
		rt_spin_lock(&(global.hard_lock));
		send_ipi_shorthand(APIC_DEST_ALLBUT, HARD_LOCK_IPI);
		while (cpu_own_irq[HARD_LOCK_IPI].dest_status != (cpu_present_map & ~global.locked_cpus));
	}

	return flags;
}

static inline void hard_unlock_all(unsigned long flags)
{
	if (global.hard_nesting > 0) {
		if (!(--global.hard_nesting)) {
			rt_spin_unlock(&(global.hard_lock));
			while (cpu_own_irq[HARD_LOCK_IPI].dest_status);
		}
	}

	rt_global_restore_flags(flags);
}


/* Dispatching interrupts, the old RTLinux S_STI revisited with the addition */
/* of system services requests and interprocessors interrupts.               */

static spinlock_t *ic_lock[NR_GLOBAL_IRQS];
static void (*ic_ack_irq[NR_GLOBAL_IRQS])    (unsigned int irq);
static void (*ic_mask_and_ack_irq[NR_GLOBAL_IRQS])   (unsigned int irq);
static void (*ic_unmask_irq[NR_GLOBAL_IRQS]) (unsigned int irq);
static void (*linux_unmask_irq[NR_GLOBAL_IRQS]) (unsigned int irq);

void rt_ack_irq(unsigned int irq)
{
	unsigned int flags;
	spinlock_t *dev_lock;
	flags = rt_spin_lock_irqsave(dev_lock = ic_lock[irq]);
	ic_ack_irq[irq](irq);
	rt_spin_unlock_irqrestore(flags, dev_lock);
}

void rt_mask_and_ack_irq(unsigned int irq)
{
	unsigned int flags;
	spinlock_t *dev_lock;
	flags = rt_spin_lock_irqsave(dev_lock = ic_lock[irq]);
	ic_mask_and_ack_irq[irq](irq);
	rt_spin_unlock_irqrestore(flags, dev_lock);
}

void rt_unmask_irq(unsigned int irq)
{
	unsigned int flags;
	spinlock_t *dev_lock;
	flags = rt_spin_lock_irqsave(dev_lock = ic_lock[irq]);
	ic_unmask_irq[irq](irq);
	rt_spin_unlock_irqrestore(flags, dev_lock);
}

void rt_startup_irq(unsigned int irq)
{
	unsigned int flags;
	spinlock_t *dev_lock;
	flags = rt_spin_lock_irqsave(dev_lock = ic_lock[irq]);
	linux_irq_desc_handler[irq]->startup(irq);
	rt_spin_unlock_irqrestore(flags, dev_lock);
}

void rt_shutdown_irq(unsigned int irq)
{
	unsigned int flags;
	spinlock_t *dev_lock;
	flags = rt_spin_lock_irqsave(dev_lock = ic_lock[irq]);
	linux_irq_desc_handler[irq]->shutdown(irq);
	rt_spin_unlock_irqrestore(flags, dev_lock);
}

void rt_enable_irq(unsigned int irq)
{
	unsigned int flags;
	spinlock_t *dev_lock;
	flags = rt_spin_lock_irqsave(dev_lock = ic_lock[irq]);
	linux_irq_desc_handler[irq]->enable(irq);
	rt_spin_unlock_irqrestore(flags, dev_lock);
}

void rt_disable_irq(unsigned int irq)
{
	unsigned int flags;
	spinlock_t *dev_lock;
	flags = rt_spin_lock_irqsave(dev_lock = ic_lock[irq]);
	linux_irq_desc_handler[irq]->disable(irq);
	rt_spin_unlock_irqrestore(flags, dev_lock);
}

static void do_nothing(unsigned int irq) { };

static void ack_8259_irq(unsigned int irq)
{
	if (irq & 8) {
		outb(0x62,0x20);
		outb(0x20,0xA0);
		return;
	}
	outb(0x20,0x20);
	return;
}

static void (*mask_apic_irq)(unsigned int irq);

static void mask_and_ack_apic_irq(unsigned int irq)
{
	mask_apic_irq(irq);
	ack_APIC_irq();
}

static void ack_apic_irq(unsigned int irq)
{
	ack_APIC_irq();
}

// The two dispatch function below should be static, they are not to avoid
// annoying warning messages at compile time.

int dispatch_global_irq(struct fill_t fill, int irq)
{
	spinlock_t *dev_lock;
	rt_spin_lock(dev_lock = ic_lock[irq]);
	if (global_irq_handler[irq]) {
		ic_ack_irq[irq](irq);
		rt_spin_unlock(dev_lock);
		global_irq_handler[irq]();
		rt_spin_lock_irq(&(global.data_lock));
	} else {
		ic_mask_and_ack_irq[irq](irq);
		rt_spin_unlock(dev_lock);
		rt_spin_lock(&(global.data_lock));
// This is a hack to avoid lock keys timeout.
		if (irq == 1 && test_bit(1, &global.activ_irqs)) {
			rt_spin_unlock_irq(&(global.data_lock));
			do_linux_irq(global_vector[1]);
			return 0;
		}
// End of hack.
		set_bit(irq, &(global.pending_irqs));
	}
	if ((global.used_by_linux & processor[hard_cpu_id()].intr_flag)) {
		rt_spin_unlock_irq(&(global.data_lock));
		linux_sti();
		return 1;
	} else {
		rt_spin_unlock(&(global.data_lock));
		return 0;
	}
}

int dispatch_cpu_own_irq(struct fill_t fill, unsigned int irq)
{
	int cpuid;
	ack_APIC_irq();
	cpuid = hard_cpu_id(); 
	if (cpu_own_irq[irq].handler) {
		cpu_own_irq[irq].handler();
		rt_spin_lock_irq(&(global.data_lock));
	} else {
		rt_spin_lock(&(global.data_lock));
		set_bit(irq, &processor[cpuid].pending_irqs);
	}
	if ((global.used_by_linux & processor[cpuid].intr_flag)) {
		rt_spin_unlock_irq(&(global.data_lock));
		linux_sti();
		return 1;
	} else {
		rt_spin_unlock(&(global.data_lock));
		return 0;
	}
}

long long dispatch_srq(unsigned int srq, unsigned int whatever)
{
	if (srq > 0 && srq < NR_GLOBAL_IRQS && sysrq[srq].user_handler) {
		return sysrq[srq].user_handler(whatever);
	}
	for (srq = 1; srq < NR_GLOBAL_IRQS; srq++) {
		if (sysrq[srq].label == whatever) {
			return (long long)srq;
		}
	}
	return 0;
}

/* Here are the trapped irq actions for Linux interrupt handlers.          */

static void trpd_startup_irq(unsigned int irq)
{
	spinlock_t *dev_lock;
	rt_spin_lock_irq(dev_lock = ic_lock[irq]);
	linux_irq_desc_handler[irq]->startup(irq);
	rt_spin_unlock_irq(dev_lock);
}

static void trpd_shutdown_irq(unsigned int irq)
{
	spinlock_t *dev_lock;
	rt_spin_lock_irq(dev_lock = ic_lock[irq]);
	linux_irq_desc_handler[irq]->shutdown(irq);
	rt_spin_unlock_irq(dev_lock);
}

static void trpd_enable_irq(unsigned int irq)
{
	spinlock_t *dev_lock;
	rt_spin_lock_irq(dev_lock = ic_lock[irq]);
	linux_irq_desc_handler[irq]->enable(irq);
	rt_spin_unlock_irq(dev_lock);
}

static void trpd_disable_irq(unsigned int irq)
{
	spinlock_t *dev_lock;
	rt_spin_lock_irq(dev_lock = ic_lock[irq]);
	linux_irq_desc_handler[irq]->disable(irq);
	rt_spin_unlock_irq(dev_lock);
}

static void trpd_handle(unsigned int irq, struct pt_regs *regs)
{
	linux_irq_desc_handler[irq]->handle(irq, regs);
}

static void trpd_mask_and_ack_irq(unsigned int irq)
{
	return;
}

static void trpd_unmask_irq(unsigned int irq)
{
	spinlock_t *dev_lock;
	rt_spin_lock_irq(dev_lock = ic_lock[irq]);
	linux_unmask_irq[irq](irq);
	rt_spin_unlock_irq(dev_lock);
}

static struct hw_interrupt_type trapped_linux_irq_type = { 
				"RT SPVISD", 
				trpd_startup_irq,
				trpd_shutdown_irq,
				trpd_handle,
				trpd_enable_irq,
				trpd_disable_irq };

static struct hw_interrupt_type real_time_irq_type = { 
				"REAL TIME", 
				do_nothing,
				do_nothing,
				trpd_handle, // in case RT irqs are shared
				do_nothing,
				do_nothing };

/* Request and free interrupts, system requests and interprocessors messages */
/* Request for regular Linux irqs also included. They are nicely chained to  */
/* Linux, forcing sharing with any already installed handler, so that we can */
/* have an echo from Linux for global handlers. We found that usefull during */
/* debug, but can be nice for a lot of other things, e.g. see the jiffies    */
/* recovery in rtai_sched.c, and the global broadcast to local apic timers.  */

static unsigned long irq_action_flags[NR_GLOBAL_IRQS];
static int chained_to_linux[NR_GLOBAL_IRQS];

int rt_request_global_irq(unsigned int irq, void (*handler)(void))
{
	unsigned long flags;

	if (irq >= NR_GLOBAL_IRQS || !handler) {
		return -EINVAL;
	}
	if (global_irq_handler[irq]) {
		return -EBUSY;
	}
	
	flags = hard_lock_all();
	IRQ_DESC[irq].handler = &real_time_irq_type;
	global_irq_handler[irq] = handler;
	linux_unmask_irq[irq] = do_nothing;
	hard_unlock_all(flags);
	return 0;
}

int rt_free_global_irq(unsigned int irq)
{
	unsigned long flags;

	if (irq >= NR_GLOBAL_IRQS || !global_irq_handler[irq]) {
		return -EINVAL;
	}

	flags = hard_lock_all();
	IRQ_DESC[irq].handler = &trapped_linux_irq_type;
	global_irq_handler[irq] = 0; 
	linux_unmask_irq[irq] = ic_unmask_irq[irq];
	hard_unlock_all(flags);
	return 0;
}

int rt_request_linux_irq(unsigned int irq,
	void (*linux_handler)(int irq, void *dev_id, struct pt_regs *regs), 
	char *linux_handler_id, void *dev_id)
{
	unsigned long flags;

	if (irq >= NR_GLOBAL_IRQS || !linux_handler) {
		return -EINVAL;
	}
	spin_lock_irqsave(rthal.irq_controller_lock, flags);
	if (!chained_to_linux[irq]++) {
		if (IRQ_DESC[irq].action) {
			irq_action_flags[irq] = IRQ_DESC[irq].action->flags;
			IRQ_DESC[irq].action->flags |= SA_SHIRQ;
		}
	}
	spin_unlock_irqrestore(rthal.irq_controller_lock, flags);
	request_irq(irq, linux_handler, SA_SHIRQ, linux_handler_id, dev_id);
	return 0;
}

int rt_free_linux_irq(unsigned int irq, void *dev_id)
{
	unsigned long flags;

	if (irq >= NR_GLOBAL_IRQS || !chained_to_linux[irq]) {
		return -EINVAL;
	}
	free_irq(irq, dev_id);
	spin_lock_irqsave(rthal.irq_controller_lock, flags);
	if (!(--chained_to_linux[irq]) && IRQ_DESC[irq].action) {
		IRQ_DESC[irq].action->flags = irq_action_flags[irq];
	}
	spin_unlock_irqrestore(rthal.irq_controller_lock, flags);
	return 0;
}

void rt_pend_linux_irq(unsigned int irq)
{
	set_bit(irq, &global.pending_irqs);
}

static struct desc_struct svdidt;

int rt_request_srq(unsigned int label, void (*rtai_handler)(void), long long (*user_handler)(unsigned int whatever))
{
	unsigned long flags;
	int srq;

	flags = rt_spin_lock_irqsave(&global.data_lock);
	if (!rtai_handler) {
		rt_spin_unlock_irqrestore(flags, &global.data_lock);
		return -EINVAL;
	}
	for (srq = 1; srq < NR_GLOBAL_IRQS; srq++) {
		if (!(sysrq[srq].rtai_handler)) {
			sysrq[srq].rtai_handler = rtai_handler;
			sysrq[srq].label = label;
			if (user_handler) {
				sysrq[srq].user_handler = user_handler;
				if (!svdidt.a && !svdidt.b) {
					svdidt = rt_set_full_intr_vect(RTAI_SYS_VECTOR, 15, 3, srqisr);
				}
			}
			rt_spin_unlock_irqrestore(flags, &global.data_lock);
			return srq;
		}
	}
	rt_spin_unlock_irqrestore(flags, &global.data_lock);
	return -EBUSY;
}

int rt_free_srq(unsigned int srq)
{
	unsigned long flags;

	flags = rt_spin_lock_irqsave(&global.data_lock);
	if (!srq || srq >= NR_GLOBAL_IRQS || !sysrq[srq].rtai_handler) {
		rt_spin_unlock_irqrestore(flags, &global.data_lock);
		return -EINVAL;
	}
	sysrq[srq].rtai_handler = 0; 
	sysrq[srq].user_handler = 0; 
	sysrq[srq].label = 0;
	for (srq = 1; srq < NR_GLOBAL_IRQS; srq++) {
		if (sysrq[srq].user_handler) {
			rt_spin_unlock_irqrestore(flags, &global.data_lock);
			return 0;
		}
	}
	if (svdidt.a || svdidt.b) {
		rt_reset_full_intr_vect(RTAI_SYS_VECTOR, svdidt);
		svdidt.a = svdidt.b = 0;
	}
	rt_spin_unlock_irqrestore(flags, &global.data_lock);
	return 0;
}

void rt_pend_linux_srq(unsigned int srq)
{
	set_bit(srq, &global.pending_srqs);
}

int rt_request_cpu_own_irq(unsigned int irq, void (*handler)(void))
{
	unsigned long flags;
	if (irq >= NR_CPU_OWN_IRQS || !handler) {
		return -EINVAL;
	}
	flags = rt_spin_lock_irqsave(&global.data_lock);
	if (cpu_own_irq[irq].handler) {
		rt_spin_unlock_irqrestore(flags, &global.data_lock);
		return -EBUSY;
	}
	cpu_own_irq[irq].dest_status = 0;
	cpu_own_irq[irq].handler = handler;
	rt_spin_unlock_irqrestore(flags, &global.data_lock);
	return 0;
}

int rt_free_cpu_own_irq(unsigned int irq)
{
	unsigned long flags;
	flags = rt_spin_lock_irqsave(&global.data_lock);
	if (irq >= NR_CPU_OWN_IRQS || !cpu_own_irq[irq].handler) {
		rt_spin_unlock_irqrestore(flags, &global.data_lock);
		return -EINVAL;
	}
	cpu_own_irq[irq].dest_status = 0;
	cpu_own_irq[irq].handler = 0;
	rt_spin_unlock_irqrestore(flags, &global.data_lock);
	return 0;
}

void rt_switch_to_linux(int cpuid)
{
	set_bit(cpuid, &global.used_by_linux);
	processor[cpuid].intr_flag = processor[cpuid].linux_intr_flag;
}

void rt_switch_to_real_time(int cpuid)
{
	processor[cpuid].linux_intr_flag = processor[cpuid].intr_flag;
	processor[cpuid].intr_flag = 0;
	clear_bit(cpuid, &global.used_by_linux);
}


/* IO_APIC entry modification code. Mostly taken from Linux io_apic.c       */

#define IO_APIC_BASE ((volatile int *)fix_to_virt(FIX_IO_APIC_BASE))

struct IO_APIC_route_entry {
	__u32	vector		:  8,
		delivery_mode	:  3,	/* 000: FIXED
					 * 001: lowest prio
					 * 111: ExtINT */
		dest_mode	:  1,	/* 0: physical, 1: logical */
		delivery_status	:  1,
		polarity	:  1,
		irr		:  1,
		trigger		:  1,	/* 0: edge, 1: level */
		mask		:  1,	/* 0: enabled, 1: disabled */
		__reserved_2	: 15;

	union {		struct { __u32
					__reserved_1	: 24,
					physical_dest	:  4,
					__reserved_2	:  4;
			} physical;

			struct { __u32
					__reserved_1	: 24,
					logical_dest	:  8;
			} logical;
	} dest;

} __attribute__ ((packed));

static inline unsigned int io_apic_read(unsigned int reg)
{
	*IO_APIC_BASE = reg;
	return *(IO_APIC_BASE+4);
}

static inline void io_apic_write(unsigned int reg, unsigned int value)
{
	*IO_APIC_BASE = reg;
	*(IO_APIC_BASE+4) = value;
}

static inline void io_apic_sync(void)
{
	(void) *(IO_APIC_BASE+4);
}

struct irq_2_pin_list { int pin, next; };

static int original_saved[NR_GLOBAL_IRQS];

static struct IO_APIC_route_entry original_entry[NR_GLOBAL_IRQS];

int rt_assign_irq_to_cpu(int irq, int cpu)
{
	struct IO_APIC_route_entry entry;
	unsigned long flags;
	int pin;

	if (smp_num_cpus > 1 && cpu >= 0 && cpu < smp_num_cpus) {	
		if (irq < 0 || irq >= NR_GLOBAL_IRQS) {
			return -EINVAL;
		}
		pin = ((struct irq_2_pin_list *)rthal.irq_2_pin + irq)->pin;
		if (pin < 0) {
			return -EINVAL;
		}
		if (test_and_set_bit(0, original_saved + irq)) {
			entry =	original_entry[irq];
		} else {
			*(((int *)&entry) + 0) = io_apic_read(0x10 + 2*pin);
			*(((int *)&entry) + 1) = io_apic_read(0x11 + 2*pin);
			original_entry[irq] = entry;
		}
		entry.delivery_mode = 0;                 // fixed.
		entry.dest_mode = 0;                     // physical.
		entry.dest.physical.physical_dest = cpu; // user assigned.
		flags = hard_lock_all();
		io_apic_write(0x10 + 2*pin, *(((int *)&entry) + 0));
		io_apic_write(0x11 + 2*pin, *(((int *)&entry) + 1));
		io_apic_sync();
		hard_unlock_all(flags);
		return 0;
	}
	return smp_num_cpus;
}

int rt_reset_irq_to_sym_mode(int irq)
{
	unsigned long flags;
	int pin;

	if (smp_num_cpus > 1 && test_and_clear_bit(0, original_saved + irq)) {
		if (irq < 0 || irq >= NR_GLOBAL_IRQS) {
			return -EINVAL;
		}
		pin = ((struct irq_2_pin_list *)rthal.irq_2_pin + irq)->pin;
		if (pin < 0) {
			return -EINVAL;
		}
		flags = hard_lock_all();
		io_apic_write(0x10 + 2*pin,
			      *(((int *)(original_entry + irq)) + 0));
		io_apic_write(0x11 + 2*pin,
                              *(((int *)(original_entry + irq)) + 1));
		io_apic_sync();
		hard_unlock_all(flags);
	}
	return 0;
}


/* RTAI mount-unmount functions to be called from the application to       */
/* initialise the real time application interface, i.e. this module, only  */
/* when it is required; so that it can stay asleep when it is not needed   */

static spinlock_t rtai_mount_lock = SPIN_LOCK_UNLOCKED;
static int rtai_mounted;

static void no_ack_apic_irq(void) { };

// Trivial, but we do things carefully, the blocking part is relatively short,
// should cause no troubles in the transition phase. 
// All the zeroings are strictly not required as mostly related to static data. 
// Done esplicitly for emphasis. Simple, just block other processors and grab 
// everything from Linux.  To this aim first block all the other cpus by using
// a dedicated HARD_LOCK_IPI and its vector without any protection.
void rt_mount_rtai(void)
{
     	unsigned int i;

	rt_spin_lock(&rtai_mount_lock);
	rtai_mounted++;
	if (rtai_mounted == 1) {

		rt_spin_lock_irq(&(global.hard_lock));
		if (smp_num_cpus > 1) {
			set_intr_vect(cpu_own_vector[HARD_LOCK_IPI], 
				      cpu_own_interrupt[HARD_LOCK_IPI]);
			cpu_own_irq[HARD_LOCK_IPI].handler = set_flat_dest_mode;
			cpu_own_irq[INVALIDATE_IPI].handler = invalidate_handler;
			send_ipi_shorthand(APIC_DEST_ALLBUT, HARD_LOCK_IPI);
// Be sure all are delivered.
			while (cpu_own_irq[HARD_LOCK_IPI].dest_status != (cpu_present_map & ~(1<<hard_cpu_id())));
		}

		rthal.disint   = linux_cli;
		rthal.enint    = linux_sti;
		rthal.getflags = linux_save_flags;
		rthal.setflags = linux_restore_flags;

		rthal.mask_and_ack_8259A = trpd_mask_and_ack_irq;
		rthal.unmask_8259A_irq   = trpd_unmask_irq;

// Re-ack or not re-ack APIC irqs? With the comment we let Linux re-ack since
// it seems to create no problem. Uncomment the line below if you want to test
// the other way.
		rthal.ack_APIC_irq   = no_ack_apic_irq;

		rthal.mask_IO_APIC_irq   = trpd_mask_and_ack_irq;
		rthal.unmask_IO_APIC_irq = trpd_unmask_irq;

		for (i = 0; i < NR_GLOBAL_IRQS; i++) {
			IRQ_DESC[i].handler = &trapped_linux_irq_type;
			set_intr_vect(global_vector[i], global_interrupt[i]); 
		}

		for (i = 0; i < NR_CPU_OWN_IRQS; i++) {
			set_intr_vect(cpu_own_vector[i], cpu_own_interrupt[i]);
		}

		rt_spin_unlock_irq(&(global.hard_lock));
// Be sure all processed.
		while (cpu_own_irq[HARD_LOCK_IPI].dest_status);
		if (smp_num_cpus > 1) {
			hard_cli();
			set_flat_dest_mode();
			cpu_own_irq[HARD_LOCK_IPI].dest_status = 0;
			cpu_own_irq[HARD_LOCK_IPI].handler = hard_lock_all_handler;
			hard_sti();
		}
		printk("\n***** RTAI NEWLY MOUNTED (MOUNT COUNT %d) ******\n\n", rtai_mounted);
	} else {
		printk("\n***** RTAI ALREADY MOUNTED (MOUNT COUNT %d) ******\n\n", rtai_mounted);
	}
	rt_spin_unlock(&rtai_mount_lock);
}

// Simple, now we can simply block other processors and copy original data back
// to Linux. The HARD_LOCK_IPI is the last one to be reset.
void rt_umount_rtai(void)
{
	int i;
	unsigned long flags;

	rt_spin_lock(&rtai_mount_lock);
	rtai_mounted--;
	if (!rtai_mounted) {

		flags = hard_lock_all();
		rthal = linux_rthal;
		for (i = 0; i < NR_GLOBAL_IRQS; i++) {
			IRQ_DESC[i].handler = linux_irq_desc_handler[i];
		}
		for (i = 0; i < 256; i++) {
			rthal.idt_table[i] = linux_idt_table[i];
		}
		cpu_own_irq[HARD_LOCK_IPI].dest_status = 0;
		cpu_own_irq[HARD_LOCK_IPI].handler = 0;
		cpu_own_irq[INVALIDATE_IPI].handler = 0;
		hard_unlock_all(flags);
		printk("\n***** RTAI UNMOUNTED (MOUNT COUNT %d) ******\n\n", rtai_mounted);
	} else {
		printk("\n***** RTAI CANNOT BE UMOUNTED (MOUNT COUNT %d) ******\n\n", rtai_mounted);
	}
	rt_spin_unlock(&rtai_mount_lock);
}


/* module init-cleanup */

// Let's prepare our side without any problem, so that there remain just a few
// things to be done when mounting RTAI. All the zeroings are strictly not 
// required as mostly related to static data. Done esplicitly for emphasis.
int init_module(void)
{
     	unsigned int i;

	linux_rthal = rthal;
	rtai_mounted = 0;
       	global.pending_irqs    = 0;
       	global.activ_irqs      = 0;
       	global.pending_srqs    = 0;
       	global.activ_srqs      = 0;
       	global.cpu_in_sti      = 0;
	global.used_by_linux   = ~(0xFFFFFFFF << smp_num_cpus);
	global.locked_cpus     = 0;
	global.hard_nesting    = 0;
      	spin_lock_init(&(global.data_lock));
      	spin_lock_init(&(global.hard_lock));
	for (i = 0; i < NR_IC_DEV_TYPES; i++) {
      		spin_lock_init(&(global.ic_lock[i]));
	}
	for (i = 0; i < NR_RT_CPUS; i++) {
		processor[i].intr_flag       = (1 << IFLAG) | (1 << i);
		processor[i].linux_intr_flag = (1 << IFLAG) | (1 << i);
		processor[i].pending_irqs    = 0;
		processor[i].activ_irqs      = 0;
	}
	for (i = 0; i < NR_CPU_OWN_IRQS; i++) {
		cpu_own_irq[i].dest_status = 0;
		cpu_own_irq[i].handler     = 0;
	}
	for (i = 0; i < 256; i++) {
		linux_idt_table[i] = rthal.idt_table[i];
		linux_isr[i] = get_intr_handler(i);
	}
	mask_apic_irq = rthal.mask_IO_APIC_irq;
	for (i = 0; i < NR_GLOBAL_IRQS; i++) {
		global_irq_handler[i] = 0;
		sysrq[i].rtai_handler = 0;
		sysrq[i].user_handler = 0;
		sysrq[i].label        = 0;
		linux_irq_desc_handler[i] = IRQ_DESC[i].handler;

		ic_mask_and_ack_irq[i] = ic_ack_irq[i] =
		linux_unmask_irq[i] = ic_unmask_irq[i] = do_nothing;
		ic_lock[i] = global.ic_lock + 2;
		if (*rthal.io_apic_irqs & (1 << i)) {
			global_vector[i] = rthal.irq_vector[i];
			ic_mask_and_ack_irq[i] = ic_ack_irq[i] = ack_apic_irq;
			ic_lock[i] = global.ic_lock + 1;
			if (IRQ_DESC[i].handler->typename[8] == 'l') {
				ic_mask_and_ack_irq[i] = mask_and_ack_apic_irq;
				linux_unmask_irq[i] =
				ic_unmask_irq[i] = rthal.unmask_IO_APIC_irq;
			}
		} else if (IRQ_DESC[i].handler->typename[0] == 'X') {
			ic_ack_irq[i] = ack_8259_irq;
			ic_mask_and_ack_irq[i] = rthal.mask_and_ack_8259A;
			linux_unmask_irq[i] =
			ic_unmask_irq[i] = rthal.unmask_8259A_irq;
			ic_lock[i] = global.ic_lock + 0;
		}
	}
// Uncomment the line below if you want to check that Linux works well when it
// is intercepted.
//	rt_mount_rtai();
	return 0;
}

static int trashed_local_timers_ipi, used_local_apic;

void cleanup_module(void)
{
// Uncomment the line below if you want to check that Linux works well when it
// is intercepted.
//	rt_umount_rtai();
	if (used_local_apic) {
		printk("***** TRASHED LOCAL APIC TIMERS IPIs: %d *****\n", trashed_local_timers_ipi);
	}
	return;
}


/*************** SOME TIMER FUNCTIONS TO BE LIKELY PUT ELSWHERE **************/

/* Real time timers. No oneshot, and related timer programming, calibration. */
/* Use the utility module. It is also to be decided if this stuff has to     */
/* stay here.                                                                */

struct calibration_data tuned;
struct rt_times rt_times;
struct rt_times rt_smp_times[NR_RT_CPUS];
static int use_local_apic;

// The two functions below are for 486s.

#define COUNTER_2_LATCH 0xFFFE
static int last_8254_counter2; 
static RTIME ts_8254;

RTIME rd_8254_ts(void)
{
	RTIME t;
	unsigned long flags;
	int inc, c2;

	rt_global_save_flags(&flags);
	rt_global_cli();
//	outb(0x80, 0x43);
	outb(0xD8, 0x43);
      	c2 = inb(0x42);
	inc = last_8254_counter2 - (c2 |= (inb(0x42) << 8));
	last_8254_counter2 = c2;
	t = (ts_8254 += (inc > 0 ? inc : inc + COUNTER_2_LATCH));
	rt_global_restore_flags(flags);
	return t;
}

void rt_setup_8254_tsc(void)
{
	unsigned long flags;
	int c;

	flags = hard_lock_all();
	outb_p(0x00, 0x43);
	c = inb_p(0x40);
	c |= inb_p(0x40) << 8;
	outb_p(0xB4, 0x43);
	outb_p(COUNTER_2_LATCH & 0xFF, 0x42);
	outb_p(COUNTER_2_LATCH >> 8, 0x42);
	ts_8254 = c + ((RTIME)LATCH)*jiffies;
	last_8254_counter2 = 0; 
	outb_p((inb_p(0x61) & 0xFD) | 1, 0x61);
	hard_unlock_all(flags);
}

static void trap_trashed_local_timers_ipi(void)
{ 
	trashed_local_timers_ipi++;
	return;
}

static void broadcast_to_local_timers(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned long temp;
	temp = (apic_read(APIC_ICR) & ~0xCDFFF) |
	       (APIC_DEST_DM_FIXED | APIC_DEST_ALLINC | LOCAL_TIMER_VECTOR);
	apic_write(APIC_ICR, temp);
} 

void rt_request_timer(void (*handler)(void), unsigned int tick, int apic)
{
#define WAIT_LOCK 5

	int count;
	unsigned long flags;
	used_local_apic = use_local_apic = apic;
	flags = hard_lock_all();
	do {
		outb(0x00, 0x43);
		count = inb(0x40);
	} while	((count | (inb(0x40) << 8)) > WAIT_LOCK);
	rt_times.tick_time = rdtsc();
	if (tick > 0) {
		rt_times.linux_tick = use_local_apic ? APIC_ICOUNT : LATCH;
		rt_times.tick_time = ((RTIME)rt_times.linux_tick)*(jiffies + 1);
		rt_times.intr_time = rt_times.tick_time + tick;
		rt_times.linux_time = rt_times.tick_time + rt_times.linux_tick;
		rt_times.periodic_tick = tick;
		if (use_local_apic) {
			global.hard_lock_all_service = 2;
			rt_free_cpu_own_irq(RTAI_APIC_TIMER_IPI);
			rt_request_cpu_own_irq(RTAI_APIC_TIMER_IPI, handler);
			rt_request_linux_irq(TIMER_8254_IRQ, broadcast_to_local_timers, "broadcast", broadcast_to_local_timers);
			setup_periodic_apic(tick, RTAI_APIC_TIMER_VECTOR);
		} else {
			outb(0x34, 0x43);
			outb(tick & 0xFF, 0x40);
			outb(tick >> 8, 0x40);
			rt_free_global_irq(TIMER_8254_IRQ);
			rt_request_global_irq(TIMER_8254_IRQ, handler);
		}
	} else {
		rt_times.linux_tick = imuldiv(LATCH, tuned.cpu_freq, FREQ_8254);
//		rt_times.tick_time = rdtsc();
		rt_times.intr_time = rt_times.tick_time + rt_times.linux_tick;
		rt_times.linux_time = rt_times.tick_time + rt_times.linux_tick;
		rt_times.periodic_tick = rt_times.linux_tick;
		if (use_local_apic) {
			global.hard_lock_all_service = 2;
			rt_free_cpu_own_irq(RTAI_APIC_TIMER_IPI);
			rt_request_cpu_own_irq(RTAI_APIC_TIMER_IPI, handler);
			rt_request_linux_irq(TIMER_8254_IRQ, broadcast_to_local_timers, "broadcast", broadcast_to_local_timers);
			setup_oneshot_apic(APIC_ICOUNT, RTAI_APIC_TIMER_VECTOR);
		} else {
			outb(0x30, 0x43);
			outb(LATCH & 0xFF, 0x40);
			outb(LATCH >> 8, 0x40);
			rt_free_global_irq(TIMER_8254_IRQ);
			rt_request_global_irq(TIMER_8254_IRQ, handler);
		}
	}
	hard_unlock_all(flags);
	return;
}

void rt_free_timer(void)
{
	unsigned long flags;
	flags = hard_lock_all();
	if (use_local_apic) {
		global.hard_lock_all_service = 3;
		rt_free_linux_irq(TIMER_8254_IRQ, broadcast_to_local_timers);
		setup_periodic_apic(APIC_ICOUNT, LOCAL_TIMER_VECTOR);
		rt_free_cpu_own_irq(RTAI_APIC_TIMER_IPI);
		rt_request_cpu_own_irq(RTAI_APIC_TIMER_IPI, trap_trashed_local_timers_ipi);
		use_local_apic = 0;
	} else {
		outb(0x34, 0x43);
		outb(LATCH & 0xFF, 0x40);
		outb(LATCH >> 8, 0x40);
		rt_free_global_irq(TIMER_8254_IRQ);
	}
	hard_unlock_all(flags);
}

void rt_request_apic_timers(void (*handler)(void), struct apic_timer_setup_data *apic_timer_data)
{
#define WAIT_LOCK 5
	int count, cpuid;
	unsigned long flags;
	struct apic_timer_setup_data *p;
	struct rt_times *rt_times;

	flags = hard_lock_all();
	global.hard_lock_all_service = 1;
	do {
		outb(0x00, 0x43);
		count = inb(0x40);
	} while	((count | (inb(0x40) << 8)) > WAIT_LOCK);
	apic_timers_sync_time = rdtsc() + imuldiv(LATCH, tuned.cpu_freq, FREQ_8254);
	for (cpuid = 0; cpuid < NR_RT_CPUS; cpuid++) {
		*(p = apic_timer_mode + cpuid) = apic_timer_data[cpuid];
		rt_times = rt_smp_times + cpuid;
		if (p->mode > 0) {
			p->mode = 1;
			rt_times->linux_tick = APIC_ICOUNT;
			rt_times->tick_time = llimd(apic_timers_sync_time, FREQ_APIC, tuned.cpu_freq);
			rt_times->periodic_tick = 
			p->count = imuldiv(p->count, FREQ_APIC, 1E9);
		} else {
			p->mode =  0;
			rt_times->linux_tick = imuldiv(LATCH, tuned.cpu_freq, FREQ_8254);
			rt_times->tick_time = apic_timers_sync_time;
			rt_times->periodic_tick = rt_times->linux_tick;
			p->count = APIC_ICOUNT;
		}
		rt_times->intr_time = rt_times->tick_time + rt_times->periodic_tick;
		rt_times->linux_time = rt_times->tick_time + rt_times->linux_tick;
	}
	if ((p = apic_timer_mode + hard_cpu_id())->mode) {
		while (rdtsc() < apic_timers_sync_time);
		setup_periodic_apic(p->count, RTAI_APIC_TIMER_VECTOR);
	} else {
		while (rdtsc() < apic_timers_sync_time);
		setup_oneshot_apic(p->count, RTAI_APIC_TIMER_VECTOR);
	}
	hard_unlock_all(flags);
	rt_free_cpu_own_irq(RTAI_APIC_TIMER_IPI);
	rt_request_cpu_own_irq(RTAI_APIC_TIMER_IPI, handler);
	rt_request_linux_irq(TIMER_8254_IRQ, broadcast_to_local_timers, "broadcast", broadcast_to_local_timers);
	for (cpuid = 0; cpuid < NR_RT_CPUS; cpuid++) {
		if ((p = apic_timer_data + cpuid)->mode > 0) {
			p->mode = 1;
			p->count = imuldiv(p->count, FREQ_APIC, 1E9);
		} else {
			p->mode = 0;
			p->count = imuldiv(p->count, tuned.cpu_freq, 1E9);
		}
	}
}

void rt_free_apic_timers(void)
{
	unsigned long flags;

	flags = hard_lock_all();
	global.hard_lock_all_service = 3;
	rt_free_linux_irq(TIMER_8254_IRQ, broadcast_to_local_timers);
	setup_periodic_apic(APIC_ICOUNT, LOCAL_TIMER_VECTOR);
	rt_free_cpu_own_irq(RTAI_APIC_TIMER_IPI);
	rt_request_cpu_own_irq(RTAI_APIC_TIMER_IPI, trap_trashed_local_timers_ipi);
	hard_unlock_all(flags);
}

int calibrate_8254(void)
{
	unsigned long flags, i;
	RTIME t;
	flags = hard_lock_all();
	outb(0x34, 0x43);
	t = rdtsc();
	for (i = 0; i < 10000; i++) { 
		outb(LATCH & 0xFF, 0x40);
		outb(LATCH >> 8, 0x40);
	}
	i = rdtsc() - t;
	hard_unlock_all(flags);
	return imuldiv(i, 100000, CPU_FREQ);
}
/********** END OF SOME TIMER FUNCTIONS TO BE LIKELY PUT ELSWHERE ***********/
